from .control import ControlBuilder
from ..lang.exceptions import Q1ValueError, Q1TypeError
from ..lang.timed_statements import AcquireStatement, AcquireWeighedStatement, AcquireTtlStatement
from .sequencer_data import (
    AcquisitionWeight, WeightCollection,
    Acquisition, AcquisitionCollection
)


class ReadoutBuilder(ControlBuilder):
    MIN_ACQUISITION_INTERVAL = 300

    def __init__(self, name, enabled_paths, max_output_voltage,
                 nco_frequency=None):
        super().__init__(name, enabled_paths, max_output_voltage, nco_frequency)
        self._acquisitions = AcquisitionCollection()
        self._weights = WeightCollection()
        self._integration_length_acq = 4
        self._thresholded_acq_rotation = 0.0
        self._thresholded_acq_threshold = 0.0
        self._trigger = None
        self._nco_prop_delay = 0
        self._ttl_acq_input_select = None
        self._ttl_acq_auto_bin_incr_en = None
        self._ttl_acq_threshold = None
        self._scope_mode = "off"

    @property
    def thresholded_acq_rotation(self):
        return self._thresholded_acq_rotation

    @thresholded_acq_rotation.setter
    def thresholded_acq_rotation(self, rotation):
        self._thresholded_acq_rotation = rotation

    @property
    def thresholded_acq_threshold(self):
        return self._thresholded_acq_threshold

    @thresholded_acq_threshold.setter
    def thresholded_acq_threshold(self, threshold):
        self._thresholded_acq_threshold = threshold

    @property
    def integration_length_acq(self):
        return self._integration_length_acq

    @integration_length_acq.setter
    def integration_length_acq(self, length):
        self._integration_length_acq = int(length)

    @property
    def nco_prop_delay(self):
        return self._nco_prop_delay

    @nco_prop_delay.setter
    def nco_prop_delay(self, delay):
        self._nco_prop_delay = delay

    @property
    def trigger(self):
        return self._trigger

    @trigger.setter
    def trigger(self, trigger):
        self._trigger = trigger

    @property
    def ttl_acq_input_select(self):
        return self._ttl_acq_input_select

    @ttl_acq_input_select.setter
    def ttl_acq_input_select(self, value):
        self._ttl_acq_input_select = value

    @property
    def ttl_acq_auto_bin_incr_en(self):
        return self._ttl_acq_auto_bin_incr_en

    @ttl_acq_auto_bin_incr_en.setter
    def ttl_acq_auto_bin_incr_en(self, enable):
        self._ttl_acq_auto_bin_incr_en = enable

    @property
    def ttl_acq_threshold(self):
        return self._ttl_acq_threshold

    @ttl_acq_threshold.setter
    def ttl_acq_threshold(self, threshold):
        self._ttl_acq_threshold = threshold

    @property
    def scope_mode(self):
        """Scope mode: 'off', 'single', or 'average'"""
        return self._scope_mode

    @scope_mode.setter
    def scope_mode(self, value: str):
        """Scope mode: 'off', 'single', or 'average'"""
        self._scope_mode = value

    def add_acquisition_bins(self, name, num_bins):
        return self._acquisitions.add_acquisition(name, num_bins)

    def add_weight(self, name, data):
        return self._weights.add_weight(name, data)

    def acquire(self, acquisition, bin_index='increment', t_offset=0):
        self.add_comment(f'acquire({acquisition}, {bin_index})')
        acquisition = self._translate_acquisition(acquisition)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        # TODO Keep track of acquisition trigger interval to prevent overruns?
        if bin_index == 'increment':
            reg_name = self._get_acquisition_reg_name(acquisition)
            bin_reg = self.Rs.init(reg_name)
            self._add_statement(AcquireStatement(t1, acquisition, bin_reg))
            self.Rs[reg_name] += 1
        else:
            self._add_statement(AcquireStatement(t1, acquisition, bin_index))

    def acquire_weighed(self, acquisition, bin_index, weight0, weight1=None, t_offset=0):
        self.add_comment(f'acquire_weighed({acquisition}, {bin_index})')
        if weight1 is None:
            weight1 = weight0
        acquisition = self._translate_acquisition(acquisition)
        weight0 = self._translate_weight(weight0)
        weight1 = self._translate_weight(weight1)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        if bin_index == 'increment':
            reg_name = self._get_acquisition_reg_name(acquisition)
            bin_reg = self.Rs.init(reg_name)
            st = AcquireWeighedStatement(t1, acquisition, bin_reg, weight0, weight1)
            self._add_statement(st)
            self.Rs[reg_name] += 1
        else:
            st = AcquireWeighedStatement(t1, acquisition, bin_index, weight0, weight1)
            self._add_statement(st)

    def acquire_ttl(self, acquisition, bin_index, enable, t_offset=0):
        self.add_comment(f'acquire_ttl({acquisition}, {bin_index}, {enable})')
        acquisition = self._translate_acquisition(acquisition)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        if bin_index == 'increment':
            if enable:
                reg_name = self._get_acquisition_reg_name(acquisition)
                bin_reg = self.Rs.init(reg_name)
                self._add_statement(AcquireTtlStatement(t1, acquisition, bin_reg, enable))
                self.Rs[reg_name] += 1
            else:
                self._add_statement(AcquireTtlStatement(t1, acquisition, 0, enable))
        else:
            self._add_statement(AcquireTtlStatement(t1, acquisition, bin_index, enable))

    def repeated_acquire(self, n, period, acquisition, bin_index='increment', t_offset=0):
        self.add_comment(f'repeated_acquire({n}, {period}, {acquisition}, {bin_index})')
        if period < ReadoutBuilder.MIN_ACQUISITION_INTERVAL:
            raise Q1ValueError(f'Acquisition period ({period} ns) too small. '
                               f'Minimum is {ReadoutBuilder.MIN_ACQUISITION_INTERVAL} ns')
        with self._local_timeline(t_offset=t_offset, duration=(n-1)*period):
            # Repeat only n-1 times to avoid wait after last acquire.
            # A wait after the last acquire could create unwanted waits in the
            # control sequencers, because acquisition is ~100 ns delayed w.r.t. control.
            if n > 1:
                with self._seq_repeat(n-1):
                    self.acquire(acquisition, bin_index)
                    self.wait(period)
            self.acquire(acquisition, bin_index)

    def acquire_frequency_sweep(self, n, period,
                                f_start, f_stop,
                                acquisition,
                                bin_index='increment',
                                acq_delay=0,
                                weight0: str | AcquisitionWeight | None = None,
                                weight1: str | AcquisitionWeight | None = None,
                                t_offset: int = 0):
        """
        Aquire `n` values with interval `period` while stepping frequency from `f_start` till `f_stop`.

        Args:
            n: number of values to acquire
            period: time between acquisitions
            f_start: start frequency
            f_stop: stop frequency, inclusive
            acq_delay: delay between changing frequency and starting acquisition
            weight0: if not None use weighed accquisition with specified weight for path 0.
            weight1: if not None use weighed accquisition with specified weight for path 1.
        Note:
            Experimentally it is determined that the nco_prop_delay should be 146 ns + delay in lines (~ 4 ns/m),
            and acq_delay should be nco_prop_delay + 4 ns.
        """
        self.add_comment(f'acquire_frequency_sweep({n}, {period}, {f_start}, {f_stop} {acquisition}, {bin_index})')
        if period < ReadoutBuilder.MIN_ACQUISITION_INTERVAL:
            raise Q1ValueError(f'Acquisition period ({period} ns) too small. '
                               f'Minimum is {ReadoutBuilder.MIN_ACQUISITION_INTERVAL} ns')
        if acq_delay > period - 20:
            raise Q1ValueError(f"acq_delay ({acq_delay}) too big. It should be less than period ({period}) - 20")

        with self._local_timeline(t_offset=t_offset, duration=(n-1)*period):
            # Repeat only n-1 times to avoid wait after last acquire.
            # A wait after the last acquire could create unwanted waits in the
            # control sequencers, because acquisition is ~100 ns delayed w.r.t. control.
            f_step = (f_stop - f_start) / (n - 1)
            self.Rs.frequency = int(f_start)
            self.set_frequency(self.Rs.frequency)
            self.wait(acq_delay)
            if n > 1:
                with self._seq_repeat(n-1):
                    if weight0 is not None or weight1 is not None:
                        self.acquire_weighed(acquisition, bin_index, weight0, weight1)
                    else:
                        self.acquire(acquisition, bin_index)
                    self.Rs.frequency += int(f_step)
                    self.wait(period - acq_delay)
                    self.set_frequency(self.Rs.frequency)
                    self.wait(acq_delay)
            if weight0 is not None or weight1 is not None:
                self.acquire_weighed(acquisition, bin_index, weight0, weight1)
            else:
                self.acquire(acquisition, bin_index)

    def repeated_acquire_weighed(self, n, period, acquisition, bin_index,
                                 weight0, weight1=None, t_offset=0):
        self.add_comment(f'repeated_acquire_weighed({n}, {period}, {acquisition}, {bin_index})')
        if period < ReadoutBuilder.MIN_ACQUISITION_INTERVAL:
            raise Q1ValueError(f'Acquisition period ({period} ns) too small. '
                               f'Minimum is {ReadoutBuilder.MIN_ACQUISITION_INTERVAL} ns')
        with self._local_timeline(t_offset=t_offset, duration=(n-1)*period):
            # Repeat only n-1 times to avoid wait after last acquire.
            # A wait after the last acquire could create unwanted waits in the
            # control sequencers, because acquisition is ~100 ns delayed w.r.t. control.
            if n > 1:
                with self._seq_repeat(n-1):
                    self.acquire_weighed(acquisition, bin_index, weight0, weight1)
                    self.wait(period)
            self.acquire_weighed(acquisition, bin_index, weight0, weight1)

    def acquire_ttl_interval(self, acquisition, bin_index, duration, t_offset=0):
        """Perform TTL acquisition for specified duration.
        """
        with self._local_timeline(t_offset=t_offset, duration=duration):
            self.acquire_ttl(acquisition, bin_index, 1)
            self.wait(duration)
            self.acquire_ttl(acquisition, bin_index, 0)

    def reset_bin_counter(self, acquisition):
        reg_name = self._get_acquisition_reg_name(acquisition)
        self.Rs[reg_name] = 0

    def _get_acquisition_reg_name(self, acquisition):
        if isinstance(acquisition, Acquisition):
            return f'_acq_{acquisition.name}'
        return f'_acq_{acquisition}'

    def _translate_acquisition(self, acquisition):
        if acquisition is None:
            return None
        if isinstance(acquisition, str):
            return self._acquisitions[acquisition]
        if isinstance(acquisition, Acquisition):
            return acquisition
        raise Q1TypeError(f'Illegal type {acquisition}')

    def _translate_weight(self, weight):
        if weight is None:
            return None
        if isinstance(weight, str):
            return self._weights[weight]
        if isinstance(weight, AcquisitionWeight):
            return weight
        raise Q1TypeError(f'Illegal type {weight}')

    def compile(self, generator, annotate=False):
        if self.scope_mode != "off":
            generator.add_acquisition(Acquisition("_scope", 1))
        super().compile(generator, annotate=annotate)





from init_pulsars import qcm0
from plot_util import plot_output

# %%
from q1pulse.instrument import Q1Instrument
instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('q1', qcm0.name, [0, 1], nco_frequency=0)
instrument.add_control('P1', qcm0.name, [2], nco_frequency=0)

# %%

p = instrument.new_program('chirp_1ns')
p.repetitions = 1

q1 = p.q1
P1 = p.P1

P1.block_pulse(7, 0.2)
q1.chirp(1000, 0.5, 10e6, 40e6)

P1.block_pulse(7, 0.2)
q1.chirp(999, 0.5, 10e6, 40e6)

P1.block_pulse(7, 0.2)
q1.chirp(998, 0.5, 10e6, 40e6)

P1.block_pulse(7, 0.2)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



from q1pulse.sequencer.sequencer_data import Wave, Acquisition, AcquisitionWeight
from q1pulse.lang.exceptions import Q1TypeError, Q1MemoryError
from q1pulse.util.q1configuration import Q1Configuration


class GeneratorData:
    def __init__(self):
        self.waveforms = {}
        self.weights = {}
        self.acquisitions = {}
        self._size_waveforms = 0
        self._size_weights = 0

    def translate_wave(self, wave: Wave):
        if not isinstance(wave, Wave):
            raise Q1TypeError(f'Unsupported type for wave: {wave}')

        waveforms = self.waveforms
        try:
            entry = waveforms[wave.name]
            return entry['index']
        except KeyError:
            index = len(waveforms)
            if index >= Q1Configuration.MAX_NUM_WAVEFORMS - 1:
                raise Q1MemoryError("Too many waveforms")
            size_waveforms = self._size_waveforms + len(wave.data)
            if size_waveforms > Q1Configuration.WAVEFORM_MEM_SIZE:
                raise Q1MemoryError("Too much waveform data for memory")
            self._size_waveforms = size_waveforms
            waveforms[wave.name] = {
                    'data': wave.data.tolist(),
                    'index': index
                    }
            return index

    def translate_acquisition(self, acquisition):
        if not isinstance(acquisition, Acquisition):
            raise Q1TypeError(f'Unsupported type for acquisition: {acquisition}')

        if acquisition.name not in self.acquisitions:
            self.acquisitions[acquisition.name] = {
                    'num_bins': acquisition.num_bins,
                    'index': len(self.acquisitions),
                    }

        return self.acquisitions[acquisition.name]['index']

    def translate_weight(self, weight):
        if not isinstance(weight, AcquisitionWeight):
            raise Q1TypeError(f'Unsupported type for weight: {weight}')

        if weight.name not in self.weights:
            index = len(self.weights)
            if index >= Q1Configuration.MAX_NUM_WEIGHTS - 1:
                raise Q1MemoryError("Too many acquisition weights")
            size_weights = self._size_weights + len(weight.data)
            if size_weights > Q1Configuration.WEIGHTS_MEM_SIZE:
                raise Q1MemoryError("Too much acquisition weight data for memory")
            self._size_weights = size_weights
            self.weights[weight.name] = {
                    'data': weight.data.tolist(),
                    'index': index,
                    }

        return self.weights[weight.name]['index']

    def get_data_dict(self):
        return {
                'waveforms': self.waveforms,
                'weights': self.weights,
                'acquisitions': self.acquisitions,
                }




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [0])
instrument.add_control('P2', qcm0.name, [1])
instrument.add_control('P3', qcm0.name, [2])

# %%

p = instrument.new_program('ramp_slow')
p.repetitions = 2

P1 = p.P1
P2 = p.P2
P3 = p.P3

c = 1

with p.parallel():
    # ramp just below stepping limit
    P1.ramp(6580, -0.001*c, 0.001*c)
    # ramp just above stepping limit
    P2.ramp(6580, -0.00101*c, 0.00101*c)
    P3.ramp(6580, -0.0011*c, 0.0011*c)

p.wait(100)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



from abc import ABC, abstractmethod
from contextlib import contextmanager


class GeneratorBase(ABC):
    @property
    @abstractmethod
    def repetitions(self):
        pass

    @repetitions.setter
    @abstractmethod
    def repetitions(self, value):
        pass

    @abstractmethod
    def start_main(self):
        pass

    @abstractmethod
    def end_main(self, time):
        pass

    @abstractmethod
    def add_comment(self, line, init_section=False):
        pass

    @abstractmethod
    def set_label(self, label):
        '''
        Label to be added to next instruction
        '''
        pass

    @abstractmethod
    def adjust_time(self, duration):
        pass

    @abstractmethod
    def block_start(self):
        pass

    @abstractmethod
    def block_end(self):
        pass

    @abstractmethod
    def jmp(self, label):
        pass

    @abstractmethod
    def jlt(self, register, n, label):
        pass

    @abstractmethod
    def jge(self, register, n, label):
        pass

    @abstractmethod
    def loop(self, register, label):
        pass

    @abstractmethod
    def move(self, source, destination, init_section=False):
        pass

    @abstractmethod
    def add(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def sub(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def asl(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def asr(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def bits_not(self, source, destination):
        pass

    @abstractmethod
    def bits_and(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def bits_or(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def bits_xor(self, lhs, rhs, destination):
        pass

    @abstractmethod
    def wait_reg(self, time, reg):
        pass

    @abstractmethod
    def set_mrk(self, time, value):
        pass

    @abstractmethod
    def awg_offset(self, time, offset0, offset1):
        pass

    @abstractmethod
    def awg_gain(self, time, gain0, gain1):
        pass

    @abstractmethod
    def reset_phase(self, time):
        pass

    @abstractmethod
    def set_freq(self, time, frequency):
        pass

    @abstractmethod
    def set_phase(self, time, phase, hires_regs):
        pass

    @abstractmethod
    def add_phase(self, time, delta, hires_regs):
        pass

    @abstractmethod
    def play(self, time, wave0, wave1):
        pass

    @abstractmethod
    def acquire(self, time, acq_index, bin_index):
        pass

    @abstractmethod
    def acquire_weighed(self, time, acq_index, bin_index, weight0, weight1):
        pass

    @abstractmethod
    def acquire_ttl(self, time, acq_index, bin_index, enable):
        pass

    @abstractmethod
    @contextmanager
    def scope(self):
        pass

    @abstractmethod
    def get_temp_reg(self):
        ''' Allocates an assembler register for temporary use within scope. '''
        pass

    @abstractmethod
    def allocate_reg(self, name):
        pass




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

# %%
instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('q1', qcm0.name, [0, 1], nco_frequency=100e6)
# add Q on channel 2, I on channel 3
instrument.add_control('q2', qcm0.name, [3, 2], nco_frequency=50e6)

p = instrument.new_program('iq_block')
p.repetitions = 1

q1 = p.q1
q2 = p.q2

q1.block_pulse(100, 0.5)
q2.block_pulse(100, 0.4)


p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



import logging
import time
from dataclasses import dataclass

from q1pulse.lang.exceptions import Q1MemoryError
from q1pulse.turbo_cluster import TurboCluster
from q1pulse.util.delayedkeyboardinterrupt import DelayedKeyboardInterrupt
from q1pulse.util.q1configuration import Q1Configuration
from q1pulse.util.qblox_version import qblox_version, Version

from .sequencer_states import translate_seq_status


logger = logging.getLogger(__name__)


@dataclass
class Sequencer:
    module_name: str
    seq_nr: int
    channels: list[int]
    max_output_voltage: float
    nco_frequency: float | None = None
    in_channels: list[int] | None = None
    label: str | None = None

    @property
    def enabled_paths(self) -> list[int]:
        n = len(self.channels)
        if n == 0:
            return []
        if n == 1:
            return [0]
        return [0, 1]


class QbloxModule:
    verbose = False
    n_sequencers = 6

    def __init__(self, pulsar):
        self.name = pulsar.name
        # check module is present in slot.
        if hasattr(pulsar, "present"):
            if not pulsar.present():
                raise Exception("No module in slot {pulsar.slot_idx}")
        self.pulsar = pulsar
        self._allocated_seq = 0
        with DelayedKeyboardInterrupt("module.__init__"):
            self.disable_all_out()
            # disable all sequencers
            for seq_nr in range(0, self.n_sequencers):
                self.enable_sync(seq_nr, False)

    @property
    def slot_idx(self):
        return self.pulsar.slot_idx

    @property
    def root_instrument(self):
        return self.pulsar.root_instrument

    def get_num_system_error(self):
        if isinstance(self.root_instrument, TurboCluster):
            return self.root_instrument.get_num_system_error(self.slot_idx)
        else:
            return 0

    def get_system_error(self):
        if isinstance(self.root_instrument, TurboCluster):
            return self.root_instrument.get_system_error(self.slot_idx)
        else:
            raise Exception("Standard Cluster has no `get_system_error` per module.")

    def get_sequencer(self, channels: list[int]) -> Sequencer:
        available_channels = list(range(self.n_channels))
        if any(ch not in available_channels for ch in channels):
            raise Exception(f"Illegal channel number(s) {channels}")
        seq_nr = self._allocate_seq_number()
        return Sequencer(self.name, seq_nr, channels, self.max_output_voltage)

    def _allocate_seq_number(self):
        if self._allocated_seq == self.n_sequencers:
            raise Exception(f"Module {self.name} is out of sequencers. "
                            f"All {self.n_sequencers} are already allocated.")
        sequencer_nr = self._allocated_seq
        self._allocated_seq += 1
        return sequencer_nr

    def disable_all_out(self):
        for seq_nr in range(0, self.n_sequencers):
            # Note: RF module connect outputs in pairs.
            n_out_ch = self.n_channels // 2 if self.pulsar.is_rf_type else self.n_channels
            for out in range(0, n_out_ch):
                self._sset(seq_nr, f"connect_out{out}", "off")

    def set_label(self, seq_nr, label):
        self.pulsar.sequencers[seq_nr].label = label

    def upload(self, seq_nr, sequence):
        if isinstance(sequence, str):
            # don't cache sequence. The contents of the file might have changed.
            filename = sequence
            # print(f"Loading {filename} to sequencer {self.pulsar.name}:{seq_nr}")
            self._sset(seq_nr, "sequence", filename, cache=False)
        else:
            for key in ["waveforms", "weights", "acquisitions"]:
                self._check_size(seq_nr, sequence, key)

            if qblox_version >= Version("0.18.0"):
                # Smart update
                self._update_sequence(seq_nr, sequence)
            else:
                self._sset(seq_nr, "sequence", sequence)

    def arm_sequencers(self):
        for seq_nr in range(0, self.n_sequencers):
            if self.enabled(seq_nr):
                self.pulsar.arm_sequencer(seq_nr)

    def start_sequencers(self):
        enabled = [self.enabled(seq_nr) for seq_nr in range(0, self.n_sequencers)]
        if any(enabled):
            self.pulsar.start_sequencer()

    def stop_sequencers(self):
        self.pulsar.stop_sequencer()

    def get_sequencer_status(self, seq_nr, timeout=0):
        status = self.pulsar.get_sequencer_status(seq_nr, timeout)
        return translate_seq_status(status)

    def enable_sync(self, seq_nr, enable):
        self._sset(seq_nr, "sync_en", enable)

    def enable_out(self, seq_nr: int, channels: list[int]) -> None:
        n = len(channels)
        if n > 0:
            self._sset(seq_nr, f"connect_out{channels[0]}", "I")
        if n == 2:
            self._sset(seq_nr, f"connect_out{channels[1]}", "Q")

    def enable_out_iq(self, seq_nr: int, channels: list[int]) -> None:
        if len(channels) == 0:
            return
        if len(channels) != 2 or channels[0] // 2 != channels[1] // 2:
            raise Exception(f"Incorrect channels {channels}. RF output must be enabled in pairs")
        ch = channels[0] // 2
        self._sset(seq_nr, f"connect_out{ch}", "IQ")

    def set_nco(self, seq_nr, nco_frequency):
        self._sset(seq_nr, "mod_en_awg", nco_frequency is not None)
        if nco_frequency is not None:
            self._sset(seq_nr, "nco_freq", nco_frequency)

    def set_mixer_gain_ratio(self, seq_nr, value):
        self._sset(seq_nr, "mixer_corr_gain_ratio", value)

    def set_mixer_phase_offset_degree(self, seq_nr, value):
        self._sset(seq_nr, "mixer_corr_phase_offset_degree", value)

    def configure_trigger_counter(self, seq_nr, address, threshold, invert):
        self._sset(seq_nr, f"trigger{address}_count_threshold", threshold)
        self._sset(seq_nr, f"trigger{address}_threshold_invert", invert)

    def set_awg_offsets(self, seq_nr, offset0, offset1):
        self._sset(seq_nr, "offset_awg_path0", offset0)
        self._sset(seq_nr, "offset_awg_path1", offset1)

    def enabled(self, seq_nr):
        seq = getattr(self.pulsar, f"sequencer{seq_nr}")
        param = seq.parameters["sync_en"]
        return param.cache()

    def disable_seq(self, sequencer):
        seq_nr = sequencer.seq_nr
        self.enable_sync(seq_nr, False)

    def enable_seq(self, sequencer):
        seq_nr = sequencer.seq_nr
        self.enable_sync(seq_nr, True)
        if self.pulsar.is_rf_type:
            self.enable_out_iq(seq_nr, sequencer.channels)
        else:
            self.enable_out(seq_nr, sequencer.channels)

    def _sset(self, seq_nr, name, value, cache=True):
        full_name = f"sequencer{seq_nr}.{name}"
        seq = getattr(self.pulsar, f"sequencer{seq_nr}")
        param = seq.parameters[name]
        try:
            if cache and param.cache.valid and param.cache() == value:
                if QbloxModule.verbose:
                    logger.debug(f"# {full_name}={value} -- cached")
                return
        except Exception:
            logger.debug(f"No cache value for {full_name}")
        param(value)
        if QbloxModule.verbose:
            logger.info(f"{full_name}={value}")

    def _mset(self, name, value, cache=True):
        """Set module parameter using cache"""
        param = self.pulsar.parameters[name]
        try:
            if cache and param.cache.valid and param.cache() == value:
                return
        except Exception:
            pass
        param(value)

    def invalidate_cache(self, seq_nr, param_name):
        seq = getattr(self.pulsar, f"sequencer{seq_nr}")
        param = seq.parameters[param_name]
        param.cache.invalidate()

    def set_out_offset(self, channel, offset_mV):
        if self.pulsar.is_rf_type:
            name = f"out{channel//2}_offset_path{channel % 2}"
            value = offset_mV
        else:
            name = f"out{channel}_offset"
            value = offset_mV/1000

        param = self.pulsar.parameters[name]
        if param.cache() == value:
            if QbloxModule.verbose:
                logger.debug(f"# {name}={value} -- cached")
            return
        param(value)

    def set_marker_invert(self, number: int, invert: bool):
        name = f"marker{number}_inv_en"
        param = self.pulsar.parameters[name]
        if param.cache() != invert:
            param(invert)

    def _update_sequence(self, seq_nr, sequence):
        seq = getattr(self.pulsar, f"sequencer{seq_nr}")
        param = seq.parameters["sequence"]
        if not param.cache.valid or param.cache() is None:
            # update all.
            self._sset(seq_nr, "sequence", sequence)
            return

        # compare sequence with cached value.
        # note: changes to loaded are immediately changed in cache.
        loaded = param.cache()

        program = sequence["program"]
        if loaded["program"] != program:
            loaded["program"] = program
            seq.update_sequence(program=program)

        for key in ["waveforms", "weights", "acquisitions"]:
            new_entries = sequence[key]
            if len(new_entries) == 0:
                continue
            self._update_sequence_part(seq, key, new_entries, loaded)
        # set the cache again. It's cleared by update_sequence
        param.cache.set(loaded)

    def _update_sequence_part(self, sequencer, key, new_entries: dict, loaded: dict):
        seq_idx = sequencer.seq_idx
        updates = {}
        loaded_entries = loaded[key]
        erase = False
        for name, entry in new_entries.items():
            if name in loaded_entries and entry == loaded_entries[name]:
                # Name, index and data match!
                if QbloxModule.verbose:
                    logger.debug(f"{self.slot_idx}.{seq_idx} Reuse {key}: {name}, index:{entry['index']}")
                continue
            else:
                index = entry["index"]
                # Avoid duplicate indices: Erase all if another loaded waveform has same index
                if any(e["index"] == index and n != name for n, e in loaded_entries.items()):
                    if QbloxModule.verbose:
                        logger.debug(f"Duplicate index; erase all {key}")
                    erase = True
                    break
                loaded_entries[name] = entry
                updates[name] = entry

        erase = erase or (self._check_size(seq_idx, loaded, key, raise_exception=False) is False)

        if QbloxModule.verbose and len(updates) and not erase:
            logger.debug(f"{self.slot_idx}.{seq_idx} Update {key}: {len(new_entries)} entries, "
                         f"write {[entry['index'] for entry in updates.values()]}")
        if erase:
            logger.debug(f"{self.slot_idx}.{seq_idx} Erase {key}: {len(new_entries)} new entries")
            # overwrite cached entries
            loaded[key] = new_entries
            sequencer.update_sequence(**{key: new_entries}, erase_existing=True)
        elif len(updates) > 0:
            # loaded entries is already updated.
            sequencer.update_sequence(**{key: updates}, erase_existing=False)

    def _check_size(self, seq_nr: int, sequence: dict, key: str, raise_exception: bool = True) -> bool:
        entries = sequence[key]
        if len(entries) == 0:
            return True

        if key == "acquisitions":
            max_bins = [131072, 131072, 131072, 65536, 65536, 63536]
            size = sum(entry["num_bins"] for entry in entries.values())
            max_size = max_bins[seq_nr]
            max_entries = max_size
        else:
            size = sum(len(entry["data"]) for entry in entries.values())
            if key == "waveforms":
                max_size = Q1Configuration.WAVEFORM_MEM_SIZE
                max_entries = Q1Configuration.MAX_NUM_WAVEFORMS
            else:
                max_size = Q1Configuration.WEIGHTS_MEM_SIZE
                max_entries = Q1Configuration.MAX_NUM_WEIGHTS
        if QbloxModule.verbose:
            logger.debug(f"{self.slot_idx}.{seq_nr} {key}: {len(entries)} entries, {size} total")
        if size > max_size:
            if raise_exception:
                raise Q1MemoryError(f"Too much {key} data ({size}) for memory of sequencer {seq_nr}")
            return False
        if len(entries) > max_entries:
            if raise_exception:
                raise Q1MemoryError(f"Too many {key} entries ({len(entries)}) for memory of sequencer {seq_nr}")
            return False
        return True


class QcmModule(QbloxModule):
    n_channels = 4

    def __init__(self, pulsar):
        super().__init__(pulsar)
        self.max_output_voltage = 2.5 if not pulsar.is_rf_type else 3.3


class QrmModule(QbloxModule):
    n_channels = 2

    def __init__(self, pulsar):
        super().__init__(pulsar)
        self.max_output_voltage = 0.5 if not pulsar.is_rf_type else 3.3
        self.disable_all_inputs()
        self._acq_ready = []

    def disable_all_inputs(self):
        for seq_nr in range(0, self.n_sequencers):
            if self.pulsar.is_rf_type:
                self._sset(seq_nr, "connect_acq", "off")
            else:
                self._sset(seq_nr, "connect_acq_I", "off")
                self._sset(seq_nr, "connect_acq_Q", "off")

    def thresholded_acq_rotation(self, seq_nr, phase_rotation):
        self._sset(seq_nr, "thresholded_acq_rotation", phase_rotation)

    def thresholded_acq_threshold(self, seq_nr, threshold):
        self._sset(seq_nr, "thresholded_acq_threshold", threshold)

    def integration_length_acq(self, seq_nr, length):
        self._sset(seq_nr, "integration_length_acq", length)

    def nco_prop_delay(self, seq_nr, delay):
        if delay > 0:
            if 96 <= delay <= 245:
                self._sset(seq_nr, "nco_prop_delay_comp_en", True)
                self._sset(seq_nr, "nco_prop_delay_comp", delay - 146)
            else:
                logger.warning(f"NCO delay ({delay} ns) is out of range. NCO delay not enabled.")
                self._sset(seq_nr, "nco_prop_delay_comp_en", False)
        else:
            self._sset(seq_nr, "nco_prop_delay_comp_en", False)

    def delete_acquisition_data(self, seq_nr):
        self.pulsar.delete_acquisition_data(seq_nr, all=True)

    def enable_in(self, seq_nr, channel):
        if self.pulsar.is_rf_type:
            self._sset(seq_nr, "connect_acq", "in0")
        else:
            # Keep old convention: I on 0, Q on 1
            # TODO: change API to allow different IQ mapping.
            if channel == 0:
                self._sset(seq_nr, "connect_acq_I", "in0")
            else:
                self._sset(seq_nr, "connect_acq_Q", "in1")

    def enable_seq(self, sequencer):
        super().enable_seq(sequencer)
        if sequencer.in_channels is not None:
            seq_nr = sequencer.seq_nr
            for ch in sequencer.in_channels:
                self.enable_in(seq_nr, ch)

    def set_nco(self, seq_nr, nco_frequency):
        super().set_nco(seq_nr, nco_frequency)
        self._sset(seq_nr, "demod_en_acq", nco_frequency is not None)

    def set_trigger(self, seq_nr, address, invert=False):
        enabled = address is not None and address > 0
        self._sset(seq_nr, "thresholded_acq_trigger_en", enabled)
        if enabled:
            self._sset(seq_nr, "thresholded_acq_trigger_address", address)
            self._sset(seq_nr, "thresholded_acq_trigger_invert", invert)

    def set_ttl(self, seq_nr, ttl_acq_input_select, ttl_acq_threshold, ttl_acq_auto_bin_incr_en):
        if ttl_acq_input_select is not None:
            self._sset(seq_nr, "ttl_acq_input_select", ttl_acq_input_select)
        if ttl_acq_threshold is not None:
            self._sset(seq_nr, "ttl_acq_threshold", ttl_acq_threshold)
        if ttl_acq_auto_bin_incr_en is not None:
            self._sset(seq_nr, "ttl_acq_auto_bin_incr_en", ttl_acq_auto_bin_incr_en)

    def arm_sequencers(self):
        self._acq_ready = []
        super().arm_sequencers()

    def start_sequencers(self):
        self._acq_ready = []
        super().start_sequencers()

    def mark_acq_ready(self, seq_nr):
        self._acq_ready.append(seq_nr)

    def get_acquisition_status(self, seq_nr: int, timeout_minutes: float = 1) -> bool:
        if seq_nr in self._acq_ready:
            return True
        logger.info(f"Acquisitions of {self.slot_idx}:{seq_nr} not ready when sequencer stopped")
        expiration_time = time.perf_counter() + timeout_minutes*60.0

        completed = False
        while not completed:
            with DelayedKeyboardInterrupt("get acquisition status"):
                completed = self.pulsar.get_acquisition_status(seq_nr, 0)
                logger.debug(f"Acquisition status {self.pulsar.name}:{seq_nr} ready={completed}")
            if time.perf_counter() > expiration_time:
                break
        return completed

    def get_acquisitions(self, seq_nr: int, acq_name: str):
        if qblox_version >= Version("0.18"):
            return self.pulsar.get_acquisitions(seq_nr, as_numpy=True)[acq_name]
        else:
            return self.pulsar.get_acquisitions(seq_nr)[acq_name]

    def set_scope_acq(self, seq_nr: int, average: bool = False):
        self._mset("scope_acq_sequencer_select", seq_nr)
        self._mset("scope_acq_trigger_mode_path0", "sequencer")
        self._mset("scope_acq_trigger_mode_path1", "sequencer")
        self._mset("scope_acq_avg_mode_en_path0", average)
        self._mset("scope_acq_avg_mode_en_path1", average)

    def store_scope_data(self, seq_nr: int, acq_name: str):
        self.pulsar.store_scope_acquisition(seq_nr, acq_name)



import os
import time
import logging
import uuid
from contextlib import contextmanager
from numbers import Number

from .lang.conditions import CounterFlags
from .lang.exceptions import Q1InternalError, Q1ValueError
from .lang.triggers import TriggerCounter, Trigger
from .lang.math_expressions import Expression
from .lang.timeline import Timeline
from .lang.registers import Registers
from .lang.register import Register
from .lang.register_statements import RegisterAssignment
from .lang.loops import RangeLoop, LinspaceLoop, ArrayLoop
from .assembler.generator import Q1asmGenerator

logger = logging.getLogger(__name__)


class Program:
    verbose = False

    def __init__(self, path=None):
        self.uuid = uuid.uuid4()
        self.sequence_builders = {}
        self.path = path
        self.R = Registers(self, local=False)
        self.repetitions = 1
        self._q1asm = {}
        self._loop_cnt = 0
        self._triggers = []
        # shared timeline for all sequencers
        self._timeline = Timeline()

    def add_sequence_builder(self, sequence_builder):
        name = sequence_builder.name
        self.sequence_builders[name] = sequence_builder
        setattr(self, name, sequence_builder)
        sequence_builder.start_sequence(self, self._timeline)

    def __getitem__(self, item):
        if item not in self.sequence_builders:
            raise Exception(f"no sequencer named {item}")
        return self.sequence_builders[item]

    def describe(self):
        for builder in self.sequence_builders.values():
            builder.describe()

    def seq_filename(self, name):
        os.makedirs(self.path, exist_ok=True)
        return os.path.join(self.path, f"q1seq_{name}.json")

    def compile(self, annotate=False, add_comments=True,
                listing=False, json=True, optimize=1):
        # store compiled sequences
        self._q1asm = {}

        start_compile = time.perf_counter()
        for builder in self.sequence_builders.values():
            g = Q1asmGenerator(add_comments=add_comments,
                               optimize=optimize)
            g.repetitions = self.repetitions
            start = time.perf_counter()
            builder.compile(g, annotate=annotate)
            end = time.perf_counter()
            d1 = (end-start)*1000
            start = end
            if listing or json:
                if self.path is None:
                    logger.warning("Set Q1Instrument.path or program.path for Q1ASM output to file.")
                    g.assemble()
                else:
                    filename = self.seq_filename(builder.name)
                    g.assemble(listing=listing, json_output=json, filename=filename)
            else:
                g.assemble()
            self._q1asm[builder.name] = g.q1asm
            end = time.perf_counter()
            d2 = (end-start)*1000
            if Program.verbose:
                logger.debug(f"compile {builder.name} {d1:5.2f} {d2:5.2f} ms")
        duration = time.perf_counter() - start_compile
        logger.debug(f"Total compilation {duration*1000:5.2f} ms")

    def q1asm(self, name):
        return self._q1asm[name]

    def _add_statement(self, statement, init_section=False):
        if not isinstance(statement, RegisterAssignment):
            raise Q1InternalError(f"Illegal statement for program {statement}")
        for builder in self.sequence_builders.values():
            builder._add_statement(statement, init_section=init_section)

    def loop_range(self, start_stop, stop=None, step=None):
        ''' range loop '''
        return self.__loop(RangeLoop(self._loop_cnt, start_stop, stop, step))

    def loop_linspace(self, start, end, n, endpoint=True):
        ''' repeat loop '''
        return self.__loop(LinspaceLoop(self._loop_cnt, start, end, n, endpoint))

    def loop_array(self, values):
        ''' array loop '''
        return self.__loop(ArrayLoop(self._loop_cnt, values))

    @contextmanager
    def __loop(self, loop):
        self._loop_cnt += 1
        for s in self.sequence_builders.values():
            s.enter_loop(loop)

        yield loop.loopvar

        for s in self.sequence_builders.values():
            s.exit_loop(loop)

    @contextmanager
    def parallel(self):
        self._timeline.disable_update()
        yield
        self._timeline.enable_update()

    @contextmanager
    def conditional(self, counters, t_offset=0, evaluation_time=0):
        for s in self.sequence_builders.values():
            s.enter_conditional(counters, t_offset=t_offset, evaluation_time=evaluation_time)
        flags = CounterFlags(self)
        yield flags
        for s in self.sequence_builders.values():
            s.exit_conditional()

    @contextmanager
    def condition(self, operator):
        for s in self.sequence_builders.values():
            s.enter_condition(operator)
        yield
        for s in self.sequence_builders.values():
            s.exit_condition()

    def configure_trigger(self, sequencer_name, invert=False):
        addr = len(self._triggers)+1
        trigger = Trigger(sequencer_name, address=addr, invert=invert)
        self._triggers.append(trigger)
        self.sequence_builders[sequencer_name].trigger = trigger
        return trigger

    def add_trigger_counter(self, trigger, threshold=1, invert=False):
        counter = TriggerCounter(trigger, threshold=threshold, invert=invert)
        for s in self.sequence_builders.values():
            s._add_trigger_counter(counter)
        return counter

    def latch_enable(self, enable, t_offset=0):
        for s in self.sequence_builders.values():
            s.latch_enable(enable, t_offset=t_offset)

    def latch_reset(self, t_offset=0):
        for s in self.sequence_builders.values():
            s.latch_reset(t_offset=t_offset)

    def add_comment(self, comment):
        for s in self.sequence_builders.values():
            s.add_comment(comment)

    def wait(self, t):
        if isinstance(t, Number):
            if t < 0:
                raise Q1ValueError("Wait time must be positive")
            self._timeline.set_pulse_end(self._timeline.current_time + t)
        else:
            for s in self.sequence_builders.values():
                s._add_reg_wait(t)

    def block_pulse(self, duration, sequencers, amplitudes, t_offset=0):
        if not isinstance(duration, (Register, Expression)):
            self._timeline.disable_update()
            for s, a in zip(sequencers, amplitudes):
                if isinstance(s, str):
                    s = self[s]
                s.block_pulse(duration, a, t_offset=t_offset)
            self._timeline.enable_update()
        else:
            if not self._timeline.is_running:
                raise Q1ValueError("Variable pulse length not possible in parallel section")
            self.set_offsets(sequencers, amplitudes, t_offset=t_offset)
            self.wait(duration)
            self.set_offsets(sequencers, [0.0]*len(sequencers), t_offset=0)

    def ramp(self, duration, sequencers, v_start, v_stop, t_offset=0):
        self._timeline.disable_update()
        for s, v1, v2 in zip(sequencers, v_start, v_stop):
            if isinstance(s, str):
                s = self[s]
            s.ramp(duration, v1, v2, t_offset=t_offset)
        self._timeline.enable_update()

    def set_offsets(self, sequencers, offsets, t_offset=0):
        self._timeline.disable_update()
        for s, offset in zip(sequencers, offsets):
            if isinstance(s, str):
                s = self[s]
            s.set_offset(offset, t_offset=t_offset)
        self._timeline.enable_update()




from q1pulse import Q1Instrument, lsr

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [0])

p = instrument.new_program('prng')
P1 = p.P1

R = P1.Rs

seed = 1
R.rnd = seed
p.wait(100)
with p.loop_range(10):
    # https://en.wikipedia.org/wiki/Xorshift
    # costs 27 clock ticks = 108 ns
    R.rnd ^= R.rnd << 13
    R.rnd ^= lsr(R.rnd, 17) # Python has no LSR / unsigned shift right
    R.rnd ^= R.rnd << 5
    P1.log('RND', R.rnd, time=True)
    P1.block_pulse(80+108, R.rnd.asfloat())


p.describe()

p.compile(annotate=True, listing=True)

instrument.run_program(p)

plot_output([qcm0])


# Python implementation of xorshift32
def rnd(x):
    x ^= (x << 13) & 0xFFFF_FFFF
    x ^= x >> 17
    x ^= (x << 5) & 0xFFFF_FFFF
    return x


x = 1
for i in range(10):
    x = rnd(x)
    print(f'{x:08X}')





from init_pulsars import qcm0
from plot_util import plot_output

# %%
from q1pulse.instrument import Q1Instrument
instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('q1', qcm0.name, [0, 1], nco_frequency=0)

# %%
p = instrument.new_program('chirp')
p.repetitions = 1

q1 = p.q1

q1.chirp(20_000, 0.5, 1e6, 10e6)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



from .base import Statement


class TimedStatement(Statement):
    def __init__(self, time):
        self.time = time


class MultiBranchStatement(TimedStatement):
    def __init__(self, time):
        super().__init__(time)
        self.branches = []


class WaitRegStatement(TimedStatement):
    def __init__(self, time, register):
        super().__init__(time)
        self.register = register

    def __repr__(self):
        return f'wait({self.register})'

    def write_instruction(self, generator):
        generator.wait_reg(self.time, self.register)


class SetMarkersStatement(TimedStatement):
    def __init__(self, time, value):
        super().__init__(time)
        self.value = value

    def __repr__(self):
        return f'set_mrk 0b{self.value:04b}'

    def write_instruction(self, generator):
        generator.set_mrk(self.time, self.value)


class PlayWaveStatement(TimedStatement):
    def __init__(self, time, wave0, wave1):
        super().__init__(time)
        self.wave0 = wave0
        self.wave1 = wave1

    def __repr__(self):
        wave0 = self.wave0.name if self.wave0 is not None else None
        wave1 = self.wave1.name if self.wave1 is not None else None
        return f'play {wave0} {wave1}'

    def write_instruction(self, generator):
        generator.play(self.time, self.wave0, self.wave1)


class SetFrequencyStatement(TimedStatement):
    def __init__(self, time, frequency):
        super().__init__(time)
        self.frequency = frequency

    def __repr__(self):
        return f'set_freq {self.frequency}'

    def write_instruction(self, generator):
        generator.set_freq(self.time, self.frequency)


class ResetPhaseStatement(TimedStatement):
    def __init__(self, time):
        super().__init__(time)

    def __repr__(self):
        return 'reset_phase'

    def write_instruction(self, generator):
        generator.reset_phase(self.time)


class ShiftPhaseStatement(TimedStatement):
    def __init__(self, time, delta, hires_regs=False):
        super().__init__(time)
        self.delta = delta
        self.hires_regs = hires_regs

    def __repr__(self):
        return f'phase_delta {self.delta}'

    def write_instruction(self, generator):
        generator.add_phase(self.time, self.delta, self.hires_regs)


class SetPhaseStatement(TimedStatement):
    def __init__(self, time, delta, hires_regs=False):
        super().__init__(time)
        self.delta = delta
        self.hires_regs = hires_regs

    def __repr__(self):
        return f'set_phase {self.delta}'

    def write_instruction(self, generator):
        generator.set_phase(self.time, self.delta, self.hires_regs)


class AwgDcOffsetStatement(TimedStatement):
    def __init__(self, time, offset0, offset1):
        super().__init__(time)
        self.offset0 = offset0
        self.offset1 = offset1

    def __repr__(self):
        return f'awg_offset = {self.offset0}, {self.offset1}'

    def write_instruction(self, generator):
        generator.awg_offset(self.time, self.offset0, self.offset1)


class AwgGainStatement(TimedStatement):
    def __init__(self, time, gain0, gain1):
        super().__init__(time)
        self.gain0 = gain0
        self.gain1 = gain1

    def __repr__(self):
        return f'awg_gain = {self.gain0}, {self.gain1}'

    def write_instruction(self, generator):
        generator.awg_gain(self.time, self.gain0, self.gain1)


class AcquireStatement(TimedStatement):
    def __init__(self, time, acquisition, bin_index):
        super().__init__(time)
        self.acquisition = acquisition
        self.bin_index = bin_index

    def __repr__(self):
        return f'acquire(bins={self.acquisition.name}, bin={self.bin_index})'

    def write_instruction(self, generator):
        generator.acquire(self.time, self.acquisition, self.bin_index)


class AcquireWeighedStatement(TimedStatement):
    def __init__(self, time, acquisition, bin_index, weight0, weight1):
        super().__init__(time)
        self.acquisition = acquisition
        self.bin_index = bin_index
        self.weight0 = weight0
        self.weight1 = weight1

    def __repr__(self):
        weight0 = self.weight0.name if self.weight0 is not None else None
        weight1 = self.weight1.name if self.weight1 is not None else None
        return (
            f'acquire_weighed(bins={self.acquisition.name}, bin={self.bin_index}, '
            f'weight0={weight0}, weight1={weight1})')

    def write_instruction(self, generator):
        generator.acquire_weighed(self.time, self.acquisition, self.bin_index,
                                  self.weight0, self.weight1)


class AcquireTtlStatement(TimedStatement):
    def __init__(self, time, acquisition, bin_index, enable):
        super().__init__(time)
        self.acquisition = acquisition
        self.bin_index = bin_index
        self.enable = enable

    def __repr__(self):
        return (
            f'acquire_ttl(bins={self.acquisition.name}, bin={self.bin_index}, '
            f'enable={self.enable})')

    def write_instruction(self, generator):
        generator.acquire_ttl(self.time, self.acquisition, self.bin_index, self.enable)



from typing import Any


def reduce_snapshot(snapshot: dict[str, Any], name: str | None = None):
    """Removes meaningless entries from snapshot"""

    if "__class__" not in snapshot:
        result = {}
        for key, value in snapshot.items():
            if isinstance(value, dict):
                value = reduce_snapshot(value, name=key)
                if not value:
                    continue
            result[key] = value
        return result
    else:
        exclude_keys = [
            "__class__",
            "full_name",
            "functions",
            "instrument",
            "instrument_name",
            "raw_value",
            "val_mapping",
            "validators",
            "vals",
            ]
        exclude_if_empty = [
            "inter_delay",
            "post_delay",
            "unit",
            ]
        exclude_if_name = [
            "name",
            "label",
            ]

        result = {}
        for key, value in snapshot.items():
            if key in exclude_keys:
                continue
            if key in exclude_if_empty and not value:
                continue
            if key in exclude_if_name and value == name:
                continue
            if isinstance(value, dict):
                value = reduce_snapshot(value, name=key)
                if not value:
                    continue
            result[key] = value
        return result




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [1])

p = instrument.new_program('csd')
P1 = p.P1
P2 = p.P2
R1 = p['R1'] # using alternative notation

N = 100
R1.add_acquisition_bins('default', N*N)

gates=[P1, P2] # alternative notation: ['P1', 'P2']
t_measure = 2000
t_acqdelay = 500
t_step = t_measure + t_acqdelay

with p.loop_linspace(-0.5, 0.5, N) as v2:
    with p.loop_linspace(-0.5, 0.5, N) as v1:
        with p.parallel():
            p.wait(t_step)
            P2.set_offset(v2)
            P1.set_offset(v1)
            R1.acquire('default', 'increment', t_offset=t_acqdelay)

P1.set_offset(0.0)
P2.set_offset(0.0)

p.compile(annotate=True, listing=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])

if hasattr(qrm1, 'print_acquisitions'):
    # get acquisition timing and counts from Q1Simulator
    qrm1.print_acquisitions()



from dataclasses import dataclass, field
from logging import DEBUG, INFO, WARNING, ERROR
import logging

logger = logging.getLogger(__name__)

_flag_map = {
    'DISARMED': ERROR,
    'FORCED STOP': ERROR,
    'SEQUENCE PROCESSOR Q1 ILLEGAL INSTRUCTION': ERROR,
    'SEQUENCE PROCESSOR RT EXEC ILLEGAL INSTRUCTION': ERROR,
    'SEQUENCE PROCESSOR RT EXEC COMMAND UNDERFLOW': ERROR,
    'AWG WAVE PLAYBACK INDEX INVALID PATH 0': ERROR,
    'AWG WAVE PLAYBACK INDEX INVALID PATH 1': ERROR,
    'ACQ WEIGHT PLAYBACK INDEX INVALID PATH 0': ERROR,
    'ACQ WEIGHT PLAYBACK INDEX INVALID PATH 1': ERROR,
    'ACQ SCOPE DONE PATH 0': DEBUG,
    'ACQ SCOPE DONE PATH 1': DEBUG,
    'ACQ SCOPE OUT OF RANGE PATH 0': WARNING,
    'ACQ SCOPE OUT OF RANGE PATH 1': WARNING,
    'ACQ SCOPE OVERWRITTEN PATH 0': INFO,
    'ACQ SCOPE OVERWRITTEN PATH 1': INFO,
    'ACQ BINNING DONE': DEBUG,
    'ACQ BINNING FIFO ERROR': ERROR,
    'ACQ BINNING COMM ERROR': ERROR,
    'ACQ BINNING OUT OF RANGE': WARNING,
    'ACQ INDEX INVALID': ERROR,
    'ACQ BIN INDEX INVALID': ERROR,
    'TRIGGER NETWORK CONFLICT': ERROR,
    'TRIGGER NETWORK MISSED INTERNAL TRIGGER': ERROR,
    'OUTPUT OVERFLOW': ERROR,
    'CLOCK INSTABILITY': ERROR,
    'ACQ INTEGRATOR OUT OF RANGE PATH 0': WARNING,
    'ACQ INTEGRATOR OUT OF RANGE PATH 1': WARNING,
    }


@dataclass
class SequencerStatus:
    state: str
    # field status added in qblox version 0.12
    status: str = 'OKAY'
    level: int = 10
    errors: list[str] = field(default_factory=list)
    warnings: list[str] = field(default_factory=list)
    info_msgs: list[str] = field(default_factory=list)
    debug_msgs: list[str] = field(default_factory=list)
    input_overloaded: bool = False

    def __post_init__(self):
        self.msg_list = {
            ERROR: self.errors,
            WARNING: self.warnings,
            INFO: self.info_msgs,
            DEBUG: self.debug_msgs,
            }

    def __str__(self):
        result = f'status:{self.status}, state:{self.state}'
        if len(self.errors):
            result += f', errors:{self.errors}'
        if len(self.warnings):
            result += f', warnings:{self.warnings}'
        if len(self.info_msgs+self.debug_msgs):
            result += f', info:{self.info_msgs+self.debug_msgs}'
        return result

    def add_flags(self, flags):
        for flag in flags:
            flag_str = str(flag).replace('_', ' ')
            if flag_str not in _flag_map:
                logger.error(f'Unknown flag {flag_str} in sequencer state')
                self.errors.append(flag_str)
                self.level = ERROR
            else:
                level = _flag_map[flag_str]
                if 'OUT OF RANGE' in flag_str:
                    self.input_overloaded = True
                self.level = max(level, self.level)
                self.msg_list[level].append(flag_str)


def translate_seq_status(seq_status):
    status = SequencerStatus(seq_status.state, seq_status.status)
    status.add_flags(seq_status.info_flags)
    status.add_flags(seq_status.warn_flags)
    status.add_flags(seq_status.err_flags)

    return status




from qblox_instruments import __version__ as qblox_version_str
from packaging.version import Version
from q1pulse import __version__ as q1pulse_version


def check_qblox_instrument_version():
    min_version = "0.17.0"
    max_version = "1.1.1"
    if qblox_version < Version(min_version):
        raise Exception(f'Q1Pulse {q1pulse_version} requires qblox_instruments {min_version}+')
    if qblox_version > Version(max_version):
        print(f'WARNING Q1Pulse {q1pulse_version} has not been tested on qblox_instruments {qblox_version}')


qblox_version = Version(qblox_version_str)



import json
import logging
import os
from datetime import datetime
from pprint import pprint

import numpy as np
from qblox_instruments import Cluster

from .reduce_snapshot import reduce_snapshot


logger = logging.getLogger(__name__)


def q1asm_dump(cluster: Cluster, path: str | None = None, sync_en_only: bool = True):
    """Writes cluster configuration including current sequences to disk.

    All data are retrieved from the qcodes cache!
    This function only works properly if called on the Cluster object
    that was used to program the cluster.

    Args:
        cluster: cluster to get data from.
        path: path to store the program.
        sync_en_only:
            If True only save sequences of sequencers with `sync_en() = True`.
            if False try to save sequences of all sequencers.
    """
    if path is None:
        path = os.getcwd()

    path = os.path.expanduser(path)
    now = datetime.now().strftime("%Y%m%d_%H%M%S_%f")

    path = os.path.join(path, f"q1program_{now}")
    print(f"dumping cluster configuration + Q1ASM to {path}")

    dumper = Q1Dumper(cluster)
    dumper.save_program(path, sync_en_only)


class Q1Dumper:

    def __init__(self, cluster: Cluster):
        self._cluster = cluster

    def save_program(self, path: str, sync_en_only: bool = True):
        """Stores program including current settings of the sequencers.

        Args:
            path: path to store the program.
            sync_en_only:
                If True only save sequences of sequencers with `sync_en() = True`.
                if False try to save sequences of all sequencers.
        """
        os.makedirs(path, exist_ok=True)

        config = {}
        cluster = self._cluster

        for slot, module in enumerate(cluster.modules, 1):
            if not module.present():
                continue
            if module.is_qcm_type:
                mod_type = "QCM"
                seq_type = "controller"
            elif module.is_qrm_type:
                mod_type = "QRM"
                seq_type = "readout"
            else:
                logger.warning(f"unknown module type {module.module_type} in slot {slot}")
                continue

            for seq_num in range(6):
                sequencer = module.sequencers[seq_num]
                name = sequencer.label
                if name == f"sequencer{seq_num}":
                    name = sequencer.name
                sync_en = sequencer.sync_en.cache()
                if sync_en_only and not sync_en:
                    continue

                filename = self._save_sequence(sequencer, name, path)

                seq_config = self._get_seq_config(sequencer, module.is_qcm_type, module.is_rf_type)

                builder_config = {
                    "instrument": cluster.name,
                    "module": module.name,
                    "slot": module.slot_idx,
                    "seq_nr": seq_num,
                    "seq_type": seq_type,
                    "sync_en": sync_en,
                    "sequence": filename,
                    }
                builder_config.update(seq_config)
                config[name] = builder_config

        with open(os.path.join(path, "q1program.json"), "w", encoding="utf-8") as f:
            json.dump(config, f, indent=2)

        # finally save snapshot. This can fail when communication to cluster is attempted.
        try:
            filename = f"snapshot_{cluster.name}.json"
            with open(os.path.join(path, filename), "w", encoding="utf-8") as f:
                snapshot = reduce_snapshot(cluster.snapshot())
                json.dump(snapshot, f, indent=1, separators=(",", ":"))
        except Exception:
            logger.error("Failed to save snapshot", exc_info=True)

    def _save_sequence(self, sequencer, name, path) -> str:
        q1asm = sequencer.sequence.cache()
        if q1asm is None:
            logger.warning(f"Q1ASM for '{name}' cannot be loaded.")
            return None
        if not isinstance(q1asm, dict):
            seq_file = q1asm
            if not os.path.exists(seq_file):
                logger.warning(f"Cannot load sequence {seq_file}")
                return None
            with open(seq_file) as fp:
                q1asm = json.load(fp)
        if q1asm is not None:
            filename = f"q1seq_{name}.json"
            with open(os.path.join(path, filename), "w", encoding="utf-8") as f:
                json.dump(q1asm, f, indent=1, separators=(",", ":"))
            self._save_q1asm(q1asm, name, path)
        else:
            filename = None
        return filename

    def _save_q1asm(self, program, name, path):
        filename = f"q1seq_{name}.q1asm"
        with open(os.path.join(path, filename), "w", encoding="utf-8") as f:
            f.write('waveforms=')
            self._pprint_data(program['waveforms'], f)
            f.write('weights=')
            self._pprint_data(program['weights'], f)
            f.write('acquisitions=')
            pprint(program['acquisitions'], f)
            f.write('\n')
            f.write('seq_prog="""\n')
            for line in program["program"].split(r"\n"):
                f.write(f"{line}\n")
            f.write('\n"""\n\n')

    def _pprint_data(self, data_dict, f):
        prefix = ' '*12
        f.write('{\n')
        for name, wave in data_dict.items():
            f.write(f"    '{name}':{{\n")
            f.write(f"        'data':\n")
            f.write(prefix)
            # precision of 5 digits is sufficient for 16 bit numbers in range [-1, +1]
            f.write(np.array2string(np.array(wave['data']),
                                    prefix=prefix,
                                    separator=',',
                                    formatter={'float_kind': lambda x: f'{x:9.5f}'},
                                    threshold=1000_000))
            f.write(f",\n")
            f.write(f"        'index':{wave['index']},\n")
            f.write(f'        }},\n')
        f.write('    }\n\n')

    def _get_seq_config(self, sequencer, is_qcm, is_rf):
        seq_config = {}
        param_names = [
                "mod_en_awg",
                "nco_freq",
                'marker_ovr_en',
                'marker_ovr_value',
                'nco_phase_offs',
                'gain_awg_path0',
                'gain_awg_path1',
                'offset_awg_path0',
                'offset_awg_path1',
                ]

        for i in range(1, 16):
            param_names += [
                f'trigger{i}_count_threshold',
                f'trigger{i}_threshold_invert',
                ]

        n_out_ch = 4 if is_qcm else 2
        if is_rf:
            n_out_ch //= 2
        for ch in range(n_out_ch):
            param_names += [f"connect_out{ch}"]

        if not is_qcm:
            if is_rf:
                param_names += ["connect_acq"]
            else:
                param_names += ["connect_acq_I", "connect_acq_Q"]

        if not is_qcm:
            param_names += [
                "demod_en_acq",
                "integration_length_acq",
                'thresholded_acq_rotation',
                'thresholded_acq_threshold',
                'thresholded_acq_trigger_en',
                'thresholded_acq_trigger_address',
                'thresholded_acq_trigger_invert',
                ]

        for param_name in param_names:
            value = sequencer.parameters[param_name].cache()
            seq_config[param_name] = value

        out_channels = set()
        in_channels = set()
        enabled_paths = set()
        for ch in range(n_out_ch):
            out = seq_config[f"connect_out{ch}"]
            if out is None:
                continue
            if out != "off":
                out_channels.add(ch)
            if "I" in out:
                enabled_paths.add(0)
            if "Q" in out:
                enabled_paths.add(1)

        if not is_qcm:
            if is_rf:
                if seq_config["connect_acq"] == "in0":
                    in_channels.add(0)
            else:
                for param in ["connect_acq_I", "connect_acq_Q"]:
                    in_ch = seq_config[param]
                    if in_ch == "in0":
                        in_channels.add(0)
                    elif in_ch == "in1":
                        in_channels.add(1)
        seq_config["in_channels"] = sorted(in_channels)
        seq_config["out_channels"] = sorted(out_channels)
        seq_config["paths"] = sorted(enabled_paths)
        seq_config["nco"] = seq_config["nco_freq"] if seq_config["mod_en_awg"] else None

        return seq_config




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])

# %%

p = instrument.new_program('bb_1ns_resolution')
p.repetitions = 3

P1 = p.P1
P2 = p.P2

p.wait(8)
for i in range(4):
    with p.parallel():
        P1.block_pulse(20, 0.5, t_offset=4)
        P2.block_pulse(20-2*i, 0.5, t_offset=4+i)
    p.wait(4)

P1.wait(7)
P1.block_pulse(15, 0.4)
p.wait(100)
for i in range(4):
    with p.parallel():
        P2.ramp(11, 0.0, 0.5)
        P2.block_pulse(12, 0.5, t_offset=11)
        P2.ramp(11, 0.5, 0.0, t_offset=23)
        P1.ramp(11, 0.0, 0.5, t_offset=i)
        P1.block_pulse(12-2*i, 0.5, t_offset=11+i)
        P1.ramp(11, 0.5, 0.0, t_offset=23-i)
    p.wait(5)

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



import time
import numpy as np

from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [1])

#%%

p = instrument.new_program('csd_ramp_pulselib')
P1 = p.P1
P2 = p.P2
R1 = p['R1'] # using alternative notation

N = 100
R1.add_acquisition_bins('default', N*N)

gates=[P1, P2] # alternative notation: ['P1', 'P2']
t_measure = 10_000
t_acqdelay = 500
t_step = t_measure + t_acqdelay

for v2 in np.linspace(-0.5, 0.5, N):
    with p.parallel():
        P2.block_pulse(t_step*N, v2)
        P1.ramp(t_step*N, -0.5, 0.5)
        R1.repeated_acquire(N, t_step,
                            'default', 'increment',
                            t_offset=t_acqdelay)

#p.describe()

start = time.perf_counter()
p.compile()
duration = time.perf_counter() - start
print(f'compilation {duration*1000:6.3f} ms')
p.compile(annotate=True, listing=True)

'''
compilation: 35 ms.
'''

#%%
start = time.perf_counter()
instrument.run_program(p)
duration = time.perf_counter() - start
print(f'execution {duration*1000:6.3f} ms')
#%%
plot_output([qcm0, qrm1])



from dataclasses import dataclass
import numpy as np

from ..lang.exceptions import Q1NameError


@dataclass
class Wave:
    name: str
    data: np.ndarray


class WaveCollection:
    def __init__(self):
        self._waves = {}

    def __getitem__(self, name):
        return self.get_wave(name)

    def get_wave(self, name):
        if name not in self._waves:
            raise Q1NameError(f'Wave {name} not defined')
        return self._waves[name]

    def add_wave(self, name, data):
        if name in self._waves:
            raise Q1NameError(f'Wave {name} already defined')
        wave = Wave(name, data)
        self._waves[name] = wave
        return wave

    def get_ramp(self, n_samples, start=0.0, stop=1.0):
        if start == 0.0 and stop == 1.0:
            name = f'_ramp_{n_samples}'
        else:
            name = f'_ramp_{n_samples}_{start*1e5:.0f}_{stop*1e5:.0f}'

        try:
            return self._waves[name]
        except KeyError:
            data = np.linspace(start, stop, n_samples, endpoint=False)
            wave = self.add_wave(name, data)
            return wave

    def get_chirp(self, n_samples, f_end, margin=0) -> tuple[Wave, Wave, float]:
        # Start with f = 0 and the delta phase / ns = 0
        # Every ns f_step/n_samples is added to delta phase / ns.
        # dp/ns = 0, 1, 2, 3, 4, ...
        # phase = 0, 1, 3, 6, 10, ...
        # phase expressed in rotations!
        dp_ns_end = f_end / 1e9
        wave_size = n_samples+margin
        dp = np.linspace(0, dp_ns_end*wave_size/n_samples, wave_size, endpoint=False)
        phase = np.cumsum(dp)
        dp_next = (phase[n_samples-1] + dp_ns_end) * 2
        nameI = f'_chirp_{n_samples}_{margin}_{f_end:.0f}_real'
        nameQ = f'_chirp_{n_samples}_{margin}_{f_end:.0f}_imag'
        try:
            return self._waves[nameI], self._waves[nameQ], dp_next
        except KeyError:
            waveI = self.add_wave(nameI, np.cos(2*np.pi*phase))
            waveQ = self.add_wave(nameQ, np.sin(2*np.pi*phase))
            # phase shift for next chirp, expressed in pi*rad!
            return waveI, waveQ, dp_next


@dataclass
class Acquisition:
    name: str
    num_bins: int


class AcquisitionCollection:
    def __init__(self):
        self._acquisitions = {}

    def __getitem__(self, name):
        return self._get_acquisition(name)

    def _get_acquisition(self, name):
        if name not in self._acquisitions:
            raise Q1NameError(f"Acquisition '{name}' not defined")
        return self._acquisitions[name]

    def add_acquisition(self, name, num_bins):
        if name in self._acquisitions:
            acquisition = self._acquisitions[name]
            if acquisition.num_bins != num_bins:
                raise Exception(f"Inconsistent number of bins in acquisition {name}: "
                                f"{num_bins} != {acquisition.num_bins}")
        else:
            acquisition = Acquisition(name, num_bins)
            self._acquisitions[name] = acquisition
        return acquisition


@dataclass
class AcquisitionWeight:
    name: str
    data: np.ndarray


class WeightCollection:
    def __init__(self):
        self._weights = {}

    def __getitem__(self, name):
        return self.get_weight(name)

    def get_weight(self, name):
        if name not in self._weights:
            raise Q1NameError(f'Weight {name} not defined')
        return self._weights[name]

    def add_weight(self, name, data):
        if name in self._weights:
            raise Q1NameError(f'Weight {name} already defined')
        weight = AcquisitionWeight(name, data)
        self._weights[name] = weight
        return weight



import numpy as np

from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

def set_mock_data(qrm, seq_nr, acq_name, data):
    seq = qrm.sequencers[seq_nr]
    if hasattr(seq, 'set_acquisition_mock_data'):
        seq.set_acquisition_mock_data([data], name=acq_name, repeat=True)

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [0])
instrument.add_readout('R1', qrm1.name, [])

# gain = 0: +/- 0.5V; gain -6 dB: +/- 1.0 V
vmax_in = 0.5
qrm1.in0_gain(0)
qrm1.in1_gain(0)

p = instrument.new_program('feedback')
p.repetitions = 2

trigger1 = p.configure_trigger('R1')
counter1 = p.add_trigger_counter(trigger1)

P1 = p.P1
R1 = p.R1

n_acq = p.repetitions
R1.add_acquisition_bins('measurements', n_acq)
R1.integration_length_acq = 100
R1.thresholded_acq_threshold = 0.1

# program
R1.acquire('measurements')
p.wait(20)
p.latch_reset()
p.wait(20)
P1.latch_enable(True)
p.wait(60)
P1.set_markers(0b1111)
P1.ramp(60, -0.5, -0.25)
p.wait(40)
P1.set_markers(0b0000)
P1.ramp(60, 0.25, 0.5)
p.wait(120)
with P1.conditional([counter1], evaluation_time=20) as flags:
    with flags.all_set():
        P1.ramp(60, 0.25, 0.25)
    with flags.none_set():
        P1.ramp(60, 0.25, 0.0)

p.wait(40)
with p.conditional([counter1]) as flags:
    with flags.all_set():
        P1.ramp(60, 0.25, 0.25)

p.wait(40)
P1.ramp(60, 0.25, 0.5)
p.wait(20)
P1.latch_enable(False)
p.wait(20)

#p.describe()
#print()

p.compile(listing=True)

#%% Set Mock data

set_mock_data(qrm1, 0, 'measurements',
              np.array([0.2, -0.2]) * R1.integration_length_acq)

#%% Run and plot
# run and get results
instrument.run_program(p)

plot_output([qcm0, qrm1])

data = instrument.get_acquisition_bins('R1', 'measurements')




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1, cluster
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [0])
instrument.add_readout('R1', qrm1.name, [], in_channels=[0])

# gain = 0: +/- 0.5V; gain -6 dB: +/- 1.0 V
vmax_in = 0.5
qrm1.in0_gain(0)
qrm1.in1_gain(0)

# %%
p = instrument.new_program('acquire_ttl')
p.repetitions = 6

P1 = p.P1
R1 = p.R1

N = 5
n_acq = max(N, p.repetitions)
R1.add_acquisition_bins('ttl', n_acq*N)
R1.ttl_acq_input_select = 0
R1.ttl_acq_auto_bin_incr_en = False
R1.ttl_acq_threshold = 0.20 / 0.5  # Threshold at 220 mV

# output range QCM (P1): +/- 2.5 V
# input range QRM (R1): +/- 0.5 V (gain = 0 dB)

v1_max = 0.2/2.5  # 200 mV
v2_max = 0.2/2.5  # 200 mV

bin_index = 0 if R1.ttl_acq_auto_bin_incr_en else "increment"

# R1.reset_bin_counter("ttl")
# Generate 5 pulses per acquisition.
# Amplitude of first pulse  of sequence increments from 0 to 200 mV.
# The pulses in the pulse train increment with steps of 50 mV from 0 to 200 mV
# First acquisition no pulse above threshold.
# Every iteration over v1 there is one more pulse above threshold.
with p.loop_linspace(0, v1_max, N) as v1:

    R1.acquire_ttl("ttl", bin_index, 1, t_offset=160)
    p.wait(10)

    with p.loop_linspace(0, v2_max, N) as v2:
        P1.block_pulse(500, v1+v2)
        p.wait(500)

    R1.acquire_ttl("ttl", 0, 0, t_offset=160)

    p.wait(2000)

# p.describe()
# print()

p.compile(listing=True, annotate=True)

# print(p.q1asm("P1"))

instrument.run_program(p)

plot_output([qcm0, qrm1])

data_ttl = instrument.get_acquisition_bins('R1', 'ttl')
print(data_ttl)

ttl_counts = data_ttl['avg_cnt']

print(ttl_counts)



import io
import json
import logging
import re
from typing import Any

_use_legacy = False

try:
    # qblox-instruments < v1.1.0
    from qblox_instruments.scpi import Cluster as ClusterScpi
    from qblox_instruments import Cluster
except ImportError:
    # Qblox-instruments v1.1+
    if _use_legacy:
        from qblox_instruments.native import ClusterLegacy as Cluster
        from qblox_instruments.scpi.layers.cluster_mm_1_0 import Cluster as ClusterScpi
    else:
        from qblox_instruments import Cluster
        from qblox_instruments.scpi import Scpi
try:
    from qblox_instruments.native.helpers import Ieee488_2Connection
    _ieee_connection_defined = True
except ImportError:
    _ieee_connection_defined = False

from qblox_instruments.ieee488_2 import Ieee488_2, IpTransport
from qblox_instruments.pnp import resolve
from q1pulse.util.qblox_version import check_qblox_instrument_version, qblox_version, Version

from qblox_instruments import (
    SequencerStatus,
    SequencerStatuses,
    SequencerStates,
    SequencerStatusFlags,
    )


logger = logging.getLogger(__name__)


class TurboCluster(Cluster):
    """
    `TurboCluster` is modified version of `qblox_instruments.Cluster` that speeds up communication
    with a factor 2 to 10.

    TurboCluster is compatible with qblox_instruments v0.11 to v0.17 and can be used as a drop-in
    replacement for Cluster when Q1Pulse is used.

    Other control code only needs some minor modifications.
    TurboCluster a TCP connection per module and therefore `get_num_system_error(slot)` should be called for
    every module. Or better, `get_system_errors()` should be called to collect the errors for
    all connections with a concurrent operation.
    The status of many sequencers can be requested with the fast concurrent call `get_sequencer_status_multiple`.

    The speedup is realized by:
        * Using a TCP/IP connection per module allowing commands to be sent to multiple modules concurrently.
        * Sending batches of commands and requests to a module and then read the batch of responses.
        * Caching module and sequencer configuration.
    """

    use_configuration_cache = True
    """If true use configuration cache. This makes configuration changes must
    faster, but could fail in scenarios where other Cluster objects can modify the configuration.
    """

    def __init__(
            self,
            name: str,
            identifier: str | None = None,
            port: int | None = None,
            debug: int | None = None,
    ):
        check_qblox_instrument_version()
        self._identifier = identifier
        addr_info = resolve(identifier)
        if addr_info.protocol != "ip":
            raise RuntimeError(
                f"Instrument cannot be reached due to invalid IP configuration. "
                f"Use qblox-pnp tool to rectify; serial number is {addr_info.address}"
            )
        self._ip_address = addr_info.address
        self._connections: dict[int | None, Ieee488_2] = {}
        self._needs_check: dict[int, bool] = {}
        self._clear_cache()
        for slot in range(1, 21):
            ip_config = resolve(f"{self._ip_address}/{slot}")
            transport = IpTransport(ip_config.address, ip_config.scpi_port, timeout=5.0)
            self._connections[slot] = Ieee488_2(transport)

        super().__init__(name, identifier, port, debug=debug)

        self._override_transport_calls()

        self._remove_slow_validators()

        # SCPI transaction map is added in v0.18 and used for commands with multiple reads like get_acquistion_data
        if hasattr(self, "_scpi_transaction_connection_map"):
            for slot, conn in self._connections.items():
                self._scpi_transaction_connection_map[slot] = Ieee488_2Connection(conn)

        # Disable continuous error checking
        # Note: Not needed anymore since v0.17.0, because default debug level has changed and
        #       start/stop sequencer has been reimplement in this class.
        self._debug = 2
        self._init_configuration_cache()

    def _remove_slow_validators(self):
        # TODO: this might not be needed anymore for version >= 0.18
        for slot in range(1, 21):
            module = self.modules[slot-1]
            if module.present():
                for seq_nr in range(6):
                    seq = module.sequencers[seq_nr]
                    # remove slow validators
                    seq.sequence._vals = []

    def _override_transport_calls(self):
        if qblox_version >= Version("1.1.0") and not _use_legacy:
            # qblox-instruments v1.1.0 Cluster has attribute _scpi.
            scpi = self._scpi
            self._connections[None] = super(Scpi, scpi)
            scpi._write = self._write
            scpi._write_bin = self._write_bin
            scpi._read = self._read
            scpi._read_bin = self._read_bin
        else:
            # Cluster version < v1.1.0 and legacy are subclasses Ieee488_2
            self._connections[None] = super(ClusterScpi, self)

    def _write(self, cmd_str):
        conn, cmd = self._get_connection_and_remove_slot(cmd_str)
        conn._write(cmd)
        # logger.debug(f"write {cmd_str}")

    def _write_bin(self, cmd_str, bin_block):
        conn, cmd = self._get_connection_and_remove_slot(cmd_str)
        conn._write_bin(cmd, bin_block)
        # logger.debug(f"write_bin {cmd_str}")

    def _read_bin(self, cmd_str, flush_line_end=True):
        conn, cmd = self._get_connection_and_remove_slot(cmd_str)
        res = conn._read_bin(cmd, flush_line_end)
        # logger.debug(f"read_bin {cmd_str}")
        return res

    def _read(self, cmd_str: str) -> str:
        conn, cmd = self._get_connection_and_remove_slot(cmd_str)
        res = conn._read(cmd)
        # logger.debug(f"read {cmd_str}")
        return res

    def _get_connection_and_remove_slot(self, cmd: str) -> tuple[Ieee488_2, str]:
        if cmd.startswith("SLOT"):
            slot_str, module_cmd = cmd.split(":", maxsplit=1)
            if len(slot_str) > 4:
                try:
                    slot = int(slot_str[4:])
                except ValueError:
                    raise Exception(f"Connect extract slot index from '{cmd}'")
                self._needs_check[slot] = True
                return self._connections[slot], module_cmd

        if None in self._connections:
            # Send to CMM without slot
            return self._connections[None], cmd
        # Cluster version < v1.1.0 and legacy are subclasses of Ieee488_2
        # There are calls to _read in __init__!
        return super(), cmd

    # ----------------------------------------------------------------
    # NOTE:
    #     arm_sequencer, start_sequencer, stop_sequencer do not need to be overridden for v0.17+
    # ----------------------------------------------------------------

    # ----------------------------------------------------------------
    # System error is a state per connection. Therefore TurboCluster has added the optional
    # slot argument to get the error count and message per connection.
    # ----------------------------------------------------------------

    def get_system_error(self, slot: int | None = None) -> str:
        """
        Get system error from queue (see `SCPI <https://www.ivifoundation.org/downloads/SCPI/scpi-99.pdf>`_).

        Parameters
        ----------
        slot : Optional[int]
            Slot number

        Returns
        -------
        str
            System error description string.

        """
        if slot is None:
            return super().get_system_error()
        else:
            return f"slot{slot}: " + self._connections[slot]._read("SYSTem:ERRor:NEXT?")

    def get_num_system_error(self, slot: int | None = None) -> int:
        """
        Get number of system errors (see `SCPI <https://www.ivifoundation.org/downloads/SCPI/scpi-99.pdf>`_).

        Parameters
        ----------
        slot : Optional[int]
            Slot number

        Returns
        -------
        int
            Current number of system errors.

        """
        if slot is None:
            cmm = self._connections.get(None, super())
            return cmm.get_num_system_error()
        else:
            if not self._needs_check.get(slot, False):
                return 0
            cnt = int(self._connections[slot]._read("SYSTem:ERRor:COUNt?"))
            if cnt == 0:
                self._needs_check[slot] = False
            return cnt

    def get_system_errors(self, exclude: list[int] = []) -> list[str]:
        """
        Returns all the system errors for all connections.
        Error counts are requested simultaneously on all connections to speed up communication.
        Error messages are requested sequentially, because there is no need to hurry when there are errors.

        Note:
            The simultaneous requests save ~1 ms per module
        """
        errors = []
        # use 0 for CMM. (actually it is in slot 0)
        slots = [0] + [slot for slot, check in self._needs_check.items() if check]
        for slot in exclude:
            try:
                slots.remove(slot)
            except Exception:
                pass
        cmm = self._connections.get(None, self)

        err_count_request = "SYSTem:ERRor:COUNt?"
        get_error = "SYSTem:ERRor:NEXT?"

        # write all requests
        for slot in slots:
            conn = self._connections.get(slot, cmm)
            conn._write(err_count_request)
        # read all responses
        for slot in slots:
            conn = self._connections.get(slot, cmm)
            # read without writing command.
            response = conn._transport.readline().rstrip()
            num_err = int(response)
            for _ in range(num_err):
                error = conn._read(get_error)
                if slot:
                    error = f"slot {slot}: " + error
                errors.append(error)
            self._needs_check[slot] = False

        return errors

    # ----------------------------------------------------------------
    # Polling the sequencer status for every sequencer in a sequential way takes a lot of time.
    # The method below sends all the status requests for all the sequencers and
    # then reads all the responses.
    # Note that the implementation uses a local implementation of `readline` that can
    # read a sequence of responses in a safe way without losing data.
    # socket.makefile creates a buffer. So all responses should be read using the
    # same buffer.
    # ----------------------------------------------------------------

    def get_sequencer_status_multiple(self, sequencers: dict[int, list[int]]) -> tuple[int, int, object]:
        """
        Returns the status for multiple sequencers using parallel requests.
        Parameters
        ----------
        sequencers : dict[int, list[int]]
            Per slot a list with sequencers to request status from.

        Note:
            The simultaneous requests save ~1 ms per sequencer
        """
        results = []
        # write all requests
        for slot, seq_nums in sequencers.items():
            if len(seq_nums) == 0:
                continue
            conn = self._connections[slot]
            for sequencer in seq_nums:
                conn._write(f"SEQuencer{sequencer}:STATE?")

        # read all responses
        for slot, seq_nums in sequencers.items():
            conn = self._connections[slot]
            filereader = conn._transport._socket.makefile()

            for sequencer in seq_nums:
                status_str = filereader.readline()
                status = _convert_sequencer_status(status_str)
                results.append((slot, sequencer, status))
            filereader.close()
        return results

    # --------------------------------------------------------------------------------
    # The following methods are added to cache the sequencer configuration and
    # reduce the amount of configuration requests every time a setting is changed.
    # --------------------------------------------------------------------------------

    def _init_configuration_cache(self):
        self._clear_cache()

    def reset(self):
        self._clear_cache()
        super().reset()

    def _clear_cache(self):
        self._channel_map_cache: dict[tuple[int, int], str] = {}
        self._sequencer_config_cache: dict[tuple[int, int], str] = {}
        self._slot_predistortion_cache: dict[int, str] = {}

    def _get_scpi(self):
        if qblox_version >= Version("1.1.0"):
            return self._scpi
        else:
            return super()
        

    def _set_sequencer_channel_map(
        self, slot: int, sequencer: int, sequencer_channel_map: Any
    ) -> None:
        """
        Set channel map of the indexed sequencer. The channel map consists of a list with two elements representing
        the components of the complex signal to be acquired, each mapping to a list of DAC indices.
        Index i maps to output i+1. If an invalid sequencer index is given or the channel map is not valid,
        an error is set in system error.

        Parameters
        ----------
        slot : int
            slot index.
        sequencer : int
            Sequencer index.
        sequencer_channel_map : Any
            Current Sequencer index Any.

        Returns
        -------
        None

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """
        if TurboCluster.use_configuration_cache:
            self._channel_map_cache[(slot, sequencer)] = json.dumps(sequencer_channel_map)
        self._get_scpi()._set_sequencer_channel_map(slot, sequencer, sequencer_channel_map)

    def _get_sequencer_channel_map(self, slot: int, sequencer: int) -> Any:
        """
        Get channel map of the indexed sequencer. The channel map consists of a list with two elements representing
        the components of the complex signal to be acquired, each mapping to a list of DAC indices.
        Index i maps to output i+1. If an invalid sequencer index is given or the channel map is not valid,
        an error is set in system error.

        Parameters
        ----------
        slot : int
            slot index.
        sequencer : int
            Sequencer index.

        Returns
        -------
        Any
            Current Sequencer index.

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """
        if TurboCluster.use_configuration_cache:
            try:
                cached_str = self._channel_map_cache[(slot, sequencer)]
                return json.loads(cached_str)
            except KeyError:
                logger.info(f"cache miss channel_map {slot}, {sequencer}")
                pass
        
            result = self._get_scpi()._get_sequencer_channel_map(slot, sequencer)
            
            if TurboCluster.use_configuration_cache:
                self._channel_map_cache[(slot, sequencer)] = json.dumps(result)
            return result
        else:
            return self._get_scpi()._get_sequencer_channel_map(slot, sequencer)

    def _set_sequencer_config(
        self, slot: int, sequencer: int, sequencer_config: Any
    ) -> None:
        """
        Set configuration of the indexed sequencer. The configuration consists of multiple parameters in a JSON format.
        If an invalid sequencer index is given, an error is set in system error.

        Parameters
        ----------
        slot : int
            slot index.
        sequencer : int
            Sequencer index.
        sequencer_config : Any
            Current Configuration struct Any.

        Returns
        -------
        None

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """

        if TurboCluster.use_configuration_cache:
            self._sequencer_config_cache[(slot, sequencer)] = json.dumps(sequencer_config)
            
        self._get_scpi()._set_sequencer_config(slot, sequencer, sequencer_config)

    def _get_sequencer_config(self, slot: int, sequencer: int) -> Any:
        """
        Get configuration of the indexed sequencer. The configuration consists of multiple parameters in a JSON format.
        If an invalid sequencer index is given, an error is set in system error.

        Parameters
        ----------
        slot : int
            slot index.
        sequencer : int
            Sequencer index.

        Returns
        -------
        Any
            Current Configuration struct.

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """
        if TurboCluster.use_configuration_cache:
            try:
                cached_str = self._sequencer_config_cache[(slot, sequencer)]
                return json.loads(cached_str)
            except KeyError:
                logger.info(f"cache miss sequencer_config {slot}, {sequencer}")
                pass
            result = self._get_scpi()._get_sequencer_config(slot, sequencer)
            self._sequencer_config_cache[(slot, sequencer)] = json.dumps(result)
            return result
        else:
            return self._get_scpi()._get_sequencer_config(slot, sequencer)

    def _set_pre_distortion_config(self, slot: int, pre_distortion_config: Any) -> None:
        """
        Set pre-distortion configuration. The configuration consists of multiple parameters in a JSON format.
        If the configation does not have the correct format, an error is set in system error..

        Parameters
        ----------
        slot : int
            slot index.
        pre_distortion_config : Any
            Current pre-distortion config Any.

        Returns
        -------
        None

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """
        if TurboCluster.use_configuration_cache:
            self._slot_predistortion_cache[slot] = json.dumps(pre_distortion_config)
        self._get_scpi()._set_pre_distortion_config(slot, pre_distortion_config)

    def _get_pre_distortion_config(self, slot: int) -> Any:
        """
        Get pre-distortion configuration. The configuration consists of multiple parameters in a JSON format.
        If the configation does not have the correct format, an error is set in system error..

        Parameters
        ----------
        slot : int
            slot index.

        Returns
        -------
        Any
            Current pre-distortion config.

        Raises
        ------
        Exception
            Invalid input parameter type.
        Exception
            An error is reported in system error and debug <= 1.
            All errors are read from system error and listed in the exception.
        """
        if TurboCluster.use_configuration_cache:
            try:
                cached_str = self._slot_predistortion_cache[slot]
                return json.loads(cached_str)
            except KeyError:
                logger.info(f"cache miss predistortion {slot}")
                pass
            result = self._get_scpi()._get_pre_distortion_config(slot)
            self._slot_predistortion_cache[slot] = json.dumps(result)
            return result
        else:
            return self._get_scpi()._get_pre_distortion_config(slot)


def readline(conn) -> str:
    """Safe implementation of readline on socket without buffering.

    The trivial solution `socket.makefile().readline()` is wrong, because
    it uses a buffer. If there is more data available on the socket, then
    some if it will be loaded into the buffer and discarded after the statement.
    Dataloss is likely with `makefile` combined with binary reads on the socket.
    """

    socket = conn._transport._socket
    buffer = io.BytesIO()
    b = None
    while True:
        b = socket.recv(1)
        buffer.write(b)
        if b == b'\n':
            break
    return buffer.getvalue().decode().rstrip()


def _convert_sequencer_status(state_str: str):
    status, state, info_flags, warn_flags, err_flags, log = _parse_sequencer_status(state_str)

    state_tuple = SequencerStatus(
        SequencerStatuses[status],
        SequencerStates[state],
        [SequencerStatusFlags[flag] for flag in info_flags],
        [SequencerStatusFlags[flag] for flag in warn_flags],
        [SequencerStatusFlags[flag] for flag in err_flags],
        log,
    )
    return state_tuple


def _parse_sequencer_status(full_status_str: str) -> tuple([list, list, list, list]):
    full_status_list = re.sub(" |-", "_", full_status_str).split(";")

    # STATUS;STATE;INFO_FLAGS;WARN_FLAGS;ERR_FLAGS;LOG
    status = full_status_list[0]  # They are always present
    state = full_status_list[1]  # They are always present

    if full_status_list[2] != "":
        info_flag_list = full_status_list[2].split(",")[:-1]
    else:
        info_flag_list = []

    if full_status_list[3] != "":
        warn_flag_list = full_status_list[3].split(",")[:-1]
    else:
        warn_flag_list = []

    if full_status_list[4] != "":
        err_flag_list = full_status_list[4].split(",")[:-1]
    else:
        err_flag_list = []

    if full_status_list[5] != "":
        log = full_status_list[5]
    else:
        log = []

    return status, state, info_flag_list, warn_flag_list, err_flag_list, log


if not _ieee_connection_defined:

    class Ieee488_2Connection:  # noqa: N801
        """
        Connection class to only expose public read/read_bin/write/write_bin methods.
        """

        def __init__(self, interface: Ieee488_2) -> None:
            if not isinstance(interface, Ieee488_2):
                raise Exception("Oops! Incompatible classes in TurboCluster")
            self._interface = interface

        def read(self, cmd_str: str) -> str:
            return self._interface._read(cmd_str)

        def read_bin(self, cmd_str: str, flush_line_end: bool = True) -> bytes:
            return self._interface._read_bin(cmd_str, flush_line_end)

        def write(self, cmd_str: str) -> None:
            self._interface._write(cmd_str)

        def write_bin(self, cmd_str: str, bin_block: bytes) -> None:
            self._interface._write_bin(cmd_str, bin_block)

        def flush_line_end(self) -> None:
            self._interface._flush_line_end()



from __future__ import annotations
import glob
import json
import logging
import os
import time
from dataclasses import dataclass
from typing import Any

from qblox_instruments import Cluster

from q1pulse.util.delayedkeyboardinterrupt import DelayedKeyboardInterrupt


logger = logging.getLogger(__name__)


def replay_program(
        cluster: Cluster,
        path: str | None = None,
        *,
        snapshot_name: str | None = None,
        module_mapping: dict[int, int] | None = None,
        sequencers: list[str] | None = None,
        sync_en_only: bool = True,
        ):
    """Replays a 'dumped' program on a cluster.

    The dumped program is a directory with snapshot and a sequence.json per sequencer.
    The program will be loaded in the cluster and played.

    Snapshot file name should match "*snapshot*.json".
    Sequence files should be named "q1seq_{name}.json" where `name` is the label of the sequencer
    if the label is not "sequencer<i>" else the sequencer name "<cluster_name>_module<slot>_sequencer<i>"
    is used.

    Args:
        cluster: Cluster (or Q1Simulator) to load the program in.
        path: directory containing the snapshot and sequence files.
        snapshot_name: filename of snapshot. Defaults to first match on "*snapshot*.json".
        module_mapping: renumbering of modules in case program is loaded on a different cluster.
        sequencers: names of sequencers to load. Defaults to all.
        sync_en_only: if True only loads and starts sequencers with `sync_en == True`.
    """
    if path is None:
        path = os.getcwd()

    path = os.path.expanduser(path)

    replayer = Replay(cluster, path,
                      module_mapping=module_mapping,
                      sequencers=sequencers,
                      sync_en_only=sync_en_only,
                      )
    replayer.read_sequences()
    replayer.load()
    replayer.play()
    replayer.check_sequencers()


class Replay:

    def __init__(
            self,
            cluster: Cluster,
            path: str,
            *,
            snapshot_name: str | None = None,
            module_mapping: dict[str, str] | None = None,
            sequencers: list[str] | None = None,
            sync_en_only: bool = True,
            ):
        self.cluster = cluster
        self.path = path

        self._module_mapping = module_mapping

        self.snapshot = self._load_snapshot(snapshot_name)

        # Ignore module settings...

        self._all_sequencers = self._get_sequencer_config()
        self._selected_sequencers = self._filter_sequencers(sequencers, sync_en_only)

    def read_sequences(self):
        for config in self._selected_sequencers:
            # find file... use label if it is specific otherwise use sequencer.name
            if config.label and config.label != f"sequencer{config.seq_num}":
                name = config.label
            else:
                name = config.name
            filename = f"q1seq_{name}.json"
            path = os.path.join(self.path, filename)
            if not os.path.exists(path):
                print("Sequence file '{filename}' not found")
            with open(path) as fp:
                config.sequence = json.load(fp)

    def load(self):
        for config in self._selected_sequencers:
            sequencer = self._get_sequencer_for_cluster(config)
            # configure sequencer
            for param_name in config.parameters_snapshot:
                if param_name == "sequence":
                    continue
                value = config.get(param_name)
                if value is not None:
                    sequencer.parameters[param_name].set(value)

            if config.sequence:
                sequencer.sequence(config.sequence)

    def play(self):
        for config in self._selected_sequencers:
            sequencer = self._get_sequencer_for_cluster(config)
            sequencer.arm_sequencer()

        self._check_system_errors()
        self.cluster.start_sequencer()
        self._check_system_errors()

    def check_sequencers(self):
        for config in self._selected_sequencers:
            sequencer = self._get_sequencer_for_cluster(config)
            status = self._get_sequencer_status(sequencer, timeout_minutes=15/60)
            print(status)

    def _load_snapshot(self, snapshot_name: str | None):
        if snapshot_name is None:
            names = glob.glob(os.path.join(self.path, "*snapshot*.json"))
            if not names:
                raise Exception("No snapshot found. Specify filename of snapshot to use.")
            filename = names[0]
            print(f"loading snapshot from {filename}")
        else:
            filename = os.path.join(self.path, snapshot_name)
        with open(filename) as fp:
            return json.load(fp)

    def _get_sequencer_config(self):
        # Assume Cluster is root of snapshot

        all_sequencers: list[SequencerConfig] = []

        # cluster_name = self.snapshot["name"]
        submodules = self.snapshot["submodules"]
        for slot_idx in range(1, 21):
            module = submodules.get(f"module{slot_idx}")
            if not module or not get_value(module["parameters"], "present"):
                continue
            module_name = module["name"]
            # module_parameters = module["parameters"]
            seq_modules = module["submodules"]
            for i in range(6):
                sequencer = seq_modules[f"sequencer{i}"]
                parameters = sequencer["parameters"]
                config = SequencerConfig(
                    name=sequencer["name"],
                    label=sequencer.get("label"),
                    module_name=module_name,
                    slot_idx=slot_idx,
                    seq_num=i,
                    parameters_snapshot=parameters,
                    )

                all_sequencers.append(config)
        return all_sequencers

    def _filter_sequencers(self, sequencers: list[str] | None, sync_en_only: bool):
        selected: list[SequencerConfig] = []

        for config in self._all_sequencers:
            sync_en = config.get("sync_en")
            if sync_en_only and not sync_en:
                continue
            if sequencers and config.name not in sequencers and config.label not in sequencers:
                continue
            selected.append(config)

        return selected

    def _get_sequencer_for_cluster(self, config: SequencerConfig):
        cluster = self.cluster
        slot_idx = config.slot_idx
        if self._module_mapping:
            slot_idx = self._module_mapping[slot_idx]
        module = cluster.modules[slot_idx-1]
        sequencer = module.sequencers[config.seq_num]
        return sequencer

    def _check_system_errors(self):
        cluster = self.cluster
        errors = []

        if hasattr(cluster, "get_system_errors"):
            # Get all errors from TurboCluster
            errors += cluster.get_system_errors()
        else:
            while cluster.get_num_system_error() != 0:
                errors.append(cluster.get_system_error())

        if len(errors) > 0:
            msg = "\n".join(errors)
            logger.error(msg)
            raise RuntimeError(msg)

    def _get_sequencer_status(self, sequencer, timeout_minutes: float):
        """Get sequencer status in a interrupt safe way.
        Only intercept the keyboard interrupt during communication,
        not when sleeping.
        """
        expiration_time = time.perf_counter() + timeout_minutes*60.0
        # timeout_poll_res = 0.001
        with DelayedKeyboardInterrupt("check status"):
            status = sequencer.get_sequencer_status(0.0)
        while (status.state == "RUNNING"
                or status.state == "Q1_STOPPED"
                or status.state == "ARMED"
               ) and time.perf_counter() < expiration_time:
            # time.sleep(timeout_poll_res)
            with DelayedKeyboardInterrupt("check status"):
                status = sequencer.get_sequencer_status(0.0)
        return status


def get_value(parameter_dict: dict[str, Any], name: str):
    return parameter_dict[name]["value"]


@dataclass
class SequencerConfig:
    name: str
    label: str
    module_name: str
    slot_idx: int
    seq_num: int
    parameters_snapshot: dict[str, Any]
    sequence: dict[str, Any] | None = None

    def get(self, name: str):
        if name not in self.parameters_snapshot:
            return None
        return self.parameters_snapshot[name]["value"]




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])

p = instrument.new_program('ramp_loop')
p.repetitions = 2

P1 = p.P1
P2 = p.P2

p.R.v_ramp1 = 0.2

P1.block_pulse(100, -0.1)
P1.block_pulse(40, 0.5)
p.wait(20)
P2.ramp(40, 0.0, p.R.v_ramp1)
P2.ramp(40, 0.2, p.R.v_ramp1)
P2.ramp(40, 0.2, 0.0)

with p.loop_linspace(-0.5, 0.5, 5) as v_ramp:
    with p.parallel():
        P2.ramp(480, 0.0, v_ramp+0.05)
        P2.ramp(140, v_ramp+0.05, 0.0, t_offset=480)
        P1.ramp(360, 0.0, v_ramp, t_offset=120)
        P1.ramp(80, v_ramp, 0.0, t_offset=480)

p.wait(100)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



from numbers import Number

from .math_expressions import get_dtype
from .register import Register
from .exceptions import Q1ValueError, Q1TypeError

class LoopVar(Register):
    def __init__(self, name, loop, **kwargs):
        super().__init__(name, **kwargs)
        self._loop = loop


class Loop:
    def __init__(self, loop_number, n, var_type=None, local=False):
        self._loop_number = loop_number
        self._n = n
        self.label = f'local_{loop_number}' if local else f'loop_{loop_number}'
        self._loop_reg = LoopVar(f'_cnt{self._loop_number}',
                                 self, local=local, dtype=int)
        if var_type:
            reg_name = f'_var{self._loop_number}'
            self._loopvar = LoopVar(reg_name, self, dtype=var_type)
        else:
            self._loopvar = None

    @property
    def n(self):
        return self._n

    @property
    def loopvar(self):
        return self._loopvar

    def __repr__(self):
        return f'repeat({self._n}):'


class RangeLoop(Loop):
    def __init__(self, loop_number, start_or_n, stop=None, step=None):
        if stop is None:
            n = start_or_n
            self._start = 0
            self._stop = n
            self._step = 1
        else:
            self._start = start_or_n
            self._stop = stop
            self._step = 1 if step is None else step
            # loop is end exclusive
            n = (self._stop-1 - self._start) // self._step
            n = max(0, n+1)

        super().__init__(loop_number, n, var_type=int)

    @property
    def start(self):
        return self._start

    @property
    def step(self):
        return self._step

    def __repr__(self):
        return f'loop_range({self._start}, {self._stop}, {self._step}):{self.loopvar}'

class LinspaceLoop(Loop):
    def __init__(self, loop_number, start, stop, n, endpoint=True):
        super().__init__(loop_number, n, var_type=float)
        if max(start,stop) > 1.0 or min(start,stop) < -1.0:
            raise Q1ValueError('value out of range [-1.0, 1.0]')
        self._start = start
        self._stop = stop
        self._endpoint = endpoint
        step_divisor = (n-1) if endpoint else n
        self._step = (stop - start)/step_divisor

    @property
    def start(self):
        return self._start

    @property
    def step(self):
        return self._step

    def __repr__(self):
        endpoint = ', endpoint=False' if not self._endpoint else ''
        return f'loop_linspace({self._start}, {self._stop}, {self.n}{endpoint}):{self.loopvar}'

class ArrayLoop(Loop):
    def __init__(self, loop_number, values):
        dtype = get_dtype(values[0])
        super().__init__(loop_number, len(values), var_type=dtype)

        self.values = values
        for value in values[1:]:
            if get_dtype(values[0]) != dtype:
                raise Q1TypeError('Array values must all be same type')
        if dtype == float:
            literal_values = [value for value in values if isinstance(value, Number)]
            if max(literal_values) > 1.0 or min(literal_values) < -1.0:
                raise Q1ValueError('value out of range [-1.0, 1.0]')

        self._table_label = f'_table{self._loop_number}'
        self._data_ptr = Register(f'_ptr{self._loop_number}', dtype=int)

    @property
    def loopvar(self):
        return self._loopvar

    def __repr__(self):
        return f'loop_array({self.values}):{self.loopvar}'



import numpy as np
from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=200e6)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [], nco_frequency=200e6)
instrument.add_control('Ro1', qrm1.name, [0,1], nco_frequency=200e6)

qrm1.in0_gain(0)
qrm1.in1_gain(0)

p = instrument.new_program('rabi')
p.repetitions = 2

q1 = p.q1
P1 = p.P1
P2 = p.P2
R1 = p['R1']
Ro1 = p['Ro1']

R1.add_acquisition_bins('default', 20)
R1.integration_length_acq = 400

gates=['P1', 'P2']
v_init = [0.220, 0.040]
v_manip = [0.0, 0.0]
v_read = [-0.050, 0.060]
rabi_amplitude = 0.1
readout_amplitude = 0.1

R1.reset_bin_counter('default')
with p.loop_range(100, 2001, 100) as t_pulse:
    #init
    p.block_pulse(200, gates, v_init)
    p.wait(20)

    # manip
    q1.block_pulse(t_pulse, rabi_amplitude)

    # read
    p.ramp(200, gates, v_manip, v_read)
    with p.parallel():
        p.set_offsets(gates, v_read)
        Ro1.block_pulse(600, readout_amplitude)
        R1.acquire('default', 'increment', t_offset=100)
        p.wait(1000)
    p.wait(1000)
p.set_offsets(gates, [0.0, 0.0])

#p.describe()

p.compile(listing=True)


instrument.run_program(p)

plot_output([qcm0, qrm1])

if hasattr(qrm1, 'print_acquisitions'):
    # get acquisition timing and counts from Q1Simulator
    qrm1.print_acquisitions()


data = instrument.get_acquisition_bins('R1', 'default')

I = data['integration']['path0']
Q = data['integration']['path1']
for i in range(20):
    c = I[i] + 1j*Q[i]
    print(f'{I[i]:6.3f}, {Q[i]:6.3f}, {abs(c):6.3f}, {np.angle(c)/np.pi:5.3f}')


import json
import os
from pprint import pprint

import numpy as np


def json2q1asm(filename: str):
    """Reformats a json file with Q1ASM sequence to a .q1asm file.

    A .q1asm file uses Python syntax to present the sequence in a readable format.
    The text can be copied to a Python script.
    """
    with open(filename) as fp:
        sequence = json.load(fp)

    if filename.endswith(".json"):
        q1asm_filename = filename[:-4] + "q1asm"
    else:
        q1asm_filename = filename + ".q1asm"

    with open(q1asm_filename, "w") as fp:
        _write_prog_and_data_txt(sequence, fp)


def _write_prog_and_data_txt(sequence, f):
    if waveforms := sequence.get('waveforms'):
        f.write('waveforms=')
        _pprint_data(waveforms, f)
    if weights := sequence.get('weights'):
        f.write('weights=')
        _pprint_data(weights, f)
    if acquisitions := sequence.get('acquisitions'):
        f.write('acquisitions=')
        pprint(acquisitions, f)
    f.write('\n')
    f.write('seq_prog="""\n')
    _pprint_q1asm_prog(sequence['program'], f)
    f.write('\n"""\n\n')


def _pprint_data(data_dict, f):
    prefix = ' '*12
    f.write('{\n')
    for name, wave in data_dict.items():
        f.write(f"    '{name}':{{\n")
        f.write(f"        'data':\n")
        f.write(prefix)
        # precision of 5 digits is sufficient for 16 bit numbers in range [-1, +1]
        f.write(np.array2string(np.array(wave['data']),
                                prefix=prefix,
                                separator=',',
                                formatter={'float_kind': lambda x: f'{x:9.5f}'},
                                threshold=1000_000))
        f.write(f",\n")
        f.write(f"        'index':{wave['index']},\n")
        f.write(f'        }},\n')
    f.write('    }\n\n')


def _pprint_q1asm_prog(program: str, f):
    lines = program.splitlines()
    for lineno, line in enumerate(lines, 1):
        if ":" in line:
            label, line = line.split(":")
            label = label + ":"
        else:
            label = ""
        line.strip(" ")
        f.write(f"{label:10} {line:40} #{lineno:04d}\n")




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])

p = instrument.new_program('phase')
p.repetitions = 100

P1 = p.P1

p.R.phase1 = 0.125
P1.Rs.phase2 = -0.125
p.wait(60)
P1.set_phase(0.5)

p.wait(348)
P1.set_phase(p.R.phase1)

p.wait(188)
P1.shift_phase(P1.Rs.phase2,hires_reg=False)

p.wait(348)
P1.shift_phase(P1.Rs.phase2 + 0.1)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)





from q1pulse import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])

p = instrument.new_program('array_loop')
p.repetitions = 2

P1 = p.P1
P2 = p.P2

P1.block_pulse(100, -0.1)
P1.block_pulse(40, 0.5)

p.wait(100)
with p.loop_array([0.1, -0.2, 0.3, -0.4, 0.5, -0.6, 0.7]) as v:
    P2.log('amplitude', v, time=True)
    P2.block_pulse(80, v)

p.wait(100)
with p.loop_array([12,80,48,56,64]) as t_wait:
    P2.set_offset(0.5)
    P1.log('t_wait', t_wait, time=True)
    p.wait(t_wait)
    P1.log('after', None, time=True)
    P1.block_pulse(20, 0.25)
    p.wait(100 - t_wait)
    P2.set_offset(0.0)
    p.wait(100)

p.wait(100)

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])
